from __future__ import absolute_import, division, print_function, unicode_literals

from amaascore.parties.organisation import Organisation


class Company(Organisation):
    """
    Represents a legal entity that has incorporated in a jurisdiction.  Does not specifiy
    anything about the structure of the company (e.g. Public, Private Limited, Partnership, etc)
    """

    def __init__(self, asset_manager_id, party_id, base_currency=None, display_name='',
                 legal_name='', url='', description='', party_status='Active',
                 year_of_incorporation=None, contact_number=None,
                 addresses=None, comments=None, emails=None, links=None, references=None,
                 *args, **kwargs):
        if not hasattr(self, 'party_class'):  # More specific child class may have already set this
            self.party_class = 'Company'
        self.year_of_incorporation = year_of_incorporation
        self.contact_number = contact_number
        super(Company, self).__init__(asset_manager_id=asset_manager_id, party_id=party_id,
                                      base_currency=base_currency, description=description,
                                      party_status=party_status, display_name=display_name,
                                      legal_name=legal_name, url=url,
                                      addresses=addresses, comments=comments, emails=emails,
                                      links=links, references=references,
                                      *args, **kwargs)



from __future__ import absolute_import, division, print_function, unicode_literals

import datetime
import uuid
import pytz

from amaascore.error_messages import ERROR_LOOKUP
from amaascore.books.enums import BOOK_TYPES
from amaascore.core.amaas_model import AMaaSModel


class Book(AMaaSModel):

    @staticmethod
    def non_interface_attributes():
        return ['positions']

    def __init__(self, asset_manager_id, book_id=None, book_type='Trading', book_status='Active', owner_id=None,
                 party_id=None, close_time=None, timezone='', base_currency='USD',
                 business_unit='', reference='', description='', positions=None,
                 *args, **kwargs):
        self.asset_manager_id = asset_manager_id
        self.book_id = book_id or uuid.uuid4().hex
        self.book_type = book_type
        self.book_status = book_status
        self.owner_id = owner_id or party_id  # This could still be None if neither are set, which will raise an error
        self.party_id = party_id
        self.close_time = close_time
        self.timezone = timezone
        self.base_currency = base_currency
        self.business_unit = business_unit
        self.reference = reference
        self.description = description
        self.positions = positions
        super(Book, self).__init__(*args, **kwargs)

    def positions_by_asset(self):
        """
        A dictionary of Position objects keyed by asset_id.
        :return:
        """
        return {position.asset_id: position for position in self.positions}

    @property
    def book_type(self):
        if hasattr(self, '_book_type'):
            return self._book_type

    @book_type.setter
    def book_type(self, book_type):
        """

        :param book_type: The type of book that we are creating - e.g. Trading, Wash
        :return:
        """
        if book_type not in BOOK_TYPES:
            raise ValueError(ERROR_LOOKUP.get('book_type_invalid') % (book_type, self.book_id, self.asset_manager_id))
        else:
            self._book_type = book_type

    def utc_book_close_time(self):
        """
        The book close time in utc.
        """
        tz = pytz.timezone(self.timezone)
        close_time = datetime.datetime.strptime(self.close_time, '%H:%M:%S').time()
        close_time = tz.localize(datetime.datetime.combine(datetime.datetime.now(tz), close_time))
        return close_time.astimezone(pytz.utc).time()



from datetime import date
from dateutil.parser import parse

from amaascore.assets.real_asset import RealAsset
from amaascore.assets.enums import *


class Automobile(RealAsset):

    def __init__(self, asset_manager_id, asset_id, client_id, asset_issuer_id=None,
                 country_id=None, display_name='', description='',
                 venue_id=None, issue_date=None, maturity_date=date.max, comments=None,
                 links=None, references=None, additional=None, currency=None,
                 asset_status='Active', model_year=None,
                 vehicle_id=None, make=None, model=None, color=None,
                 style=None, genre=None, rarity=None,
                 condition=None, imported=None, imported_country_id=None,
                 engine_id=None, chassis_id=None, steering=None, gearbox=None, gears=None,
                 drive=None, engine_cap=None, power=None, weight=None,
                 speed=None, fuel_type=None, petrol_grade=None, *args, **kwargs):
        super(Automobile, self).__init__(asset_manager_id=asset_manager_id,
                                         asset_id=asset_id, client_id=client_id,
                                         asset_issuer_id=asset_issuer_id, asset_status=asset_status,
                                         display_name=display_name,
                                         description=description, country_id=country_id,
                                         venue_id=venue_id, issue_date=issue_date,
                                         currency=currency, maturity_date=maturity_date,
                                         comments=comments, links=links, references=references,
                                         additional=additional,
                                         *args, **kwargs)
        self.model_year = model_year
        self.vehicle_id = vehicle_id
        self.make = make
        self.model = model
        self.color = color
        self.style = style
        self.genre = genre
        self.rarity = rarity
        self.condition = condition
        self.imported = imported
        self.imported_country_id = imported_country_id
        self.engine_id = engine_id
        self.chassis_id = chassis_id
        self.steering = steering
        self.gearbox = gearbox
        self.gears = gears
        self.drive = drive
        self.engine_cap = engine_cap
        self.power = power
        self.weight = weight
        self.speed = speed
        self.fuel_type = fuel_type
        self.petrol_grade = petrol_grade

    @property
    def model_year(self):
        return self._model_year

    @model_year.setter
    def model_year(self, model_year):
        if isinstance(model_year, str):
            model_year = int(model_year)
        self._model_year = model_year

    @property
    def make(self):
        return self._make

    @make.setter
    def make(self, make):
        if not make:
            self._make = None
        elif make in CAR_MAKE_RECORD:
            self._make = make
        else:
            raise ValueError("Invalid input for car make: %s" % make)

    @property
    def color(self):
        return self._color

    @color.setter
    def color(self, color):
        if not color:
            self._color = None
        elif color in CAR_COLOR_RECORD:
            self._color = color
        else:
            raise ValueError("Invalid input for car color: %s" % color)

    @property
    def style(self):
        return self._style

    @style.setter
    def style(self, style):
        if not style:
            self._style = None
        elif style in set(CAR_STYLE.keys()):
            self._style = style
        elif style in set(CAR_STYLE.values()):
            self._style = list(CAR_STYLE.keys())[list(CAR_STYLE.values()).index(style)]
        else:
            raise ValueError("Invalid input for car style: %s" % style)

    @property
    def genre(self):
        return self._genre

    @genre.setter
    def genre(self, genre):
        if not genre:
            self._genre = None
        elif genre in set(CAR_GENRE.keys()):
            self._genre = genre
        elif genre in set(CAR_GENRE.values()):
            self._genre = list(CAR_GENRE.keys())[list(CAR_GENRE.values()).index(genre)]
        else:
            raise ValueError("Invalid input for car genre: %s" % genre)

    @property
    def rarity(self):
        return self._rarity

    @rarity.setter
    def rarity(self, rarity):
        if not rarity:
            self._rarity = None
        elif rarity in CAR_RARITY:
            self._rarity = rarity
        else:
            raise ValueError("Invalid input for car rarity: %s" % rarity)

    @property
    def condition(self):
        return self._condition

    @condition.setter
    def condition(self, condition):
        if not condition:
            self._condition = None
        elif condition in set(CAR_CONDITION.keys()):
            self._condition = condition
        elif condition in set(CAR_CONDITION.values()):
            self._condition = list(CAR_CONDITION.keys())[list(CAR_CONDITION.values()).index(condition)]
        else:
            raise ValueError("Invalid input for car condition: %s" % condition)

    @property
    def imported(self):
        return self._imported

    @imported.setter
    def imported(self, imported):
        if not imported:
            self._imported = None
        else:
            if imported not in [True, False]:
                raise ValueError("imported should be True or False: %s" % imported)
            self._imported = imported

    @property
    def steering(self):
        return self._steering

    @steering.setter
    def steering(self, steering):
        if not steering:
            self._steering = None
        else:
            if steering not in ['LHD', 'RHD']:
                raise ValueError("Invalid steering LHD/RHD: %s" % steering)
            self._steering = steering

    @property
    def gearbox(self):
        return self._gearbox

    @gearbox.setter
    def gearbox(self, gearbox):
        if not gearbox:
            self._gearbox = None
        else:
            if gearbox not in ['A', 'M']:
                raise ValueError("Invalid gearbox A/M: %s" % gearbox)
            self._gearbox = gearbox

    @property
    def drive(self):
        return self._drive

    @drive.setter
    def drive(self, drive):
        if not drive:
            self._drive = None
        else:
            if drive not in ['RWD', 'FWD', 'AWD']:
                raise ValueError("Invalid drive RWD/FWD/AWD: %s" % drive)
            self._drive = drive

    @property
    def fuel_type(self):
        return self._fuel_type

    @fuel_type.setter
    def fuel_type(self, fuel_type):
        if not fuel_type:
            self._fuel_type = None
        else:
            if fuel_type not in ['Petrol', 'Diesel']:
                raise ValueError("Invalid fuel_type Petrol/Diesel: %s" % fuel_type)
            self._fuel_type = fuel_type

    @property
    def petrol_grade(self):
        return self._petrol_grade

    @petrol_grade.setter
    def petrol_grade(self, petrol_grade):
        if not petrol_grade:
            self._petrol_grade = None
        else:
            if isinstance(petrol_grade, str):
                petrol_grade = int(float(petrol_grade))
            if petrol_grade in set(CAR_PETROL_GRADE.keys()):
                self._petrol_grade = CAR_PETROL_GRADE[petrol_grade]
            else:
                raise ValueError("Invalid data type for petrol_grade (87~94): %s" % petrol_grade)



from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import json

from amaascore.config import ENVIRONMENT
from amaascore.core.amaas_model import json_handler
from amaascore.core.interface import Interface
from amaascore.transactions.utils import json_to_transaction, json_to_position, \
    json_to_mtm_result, json_to_transaction_pnl, json_to_position_pnl


class TransactionsInterface(Interface):

    def __init__(self, environment=ENVIRONMENT, logger=None, endpoint=None, username=None, password=None):
        self.logger = logger or logging.getLogger(__name__)
        super(TransactionsInterface, self).__init__(endpoint=endpoint, endpoint_type='transactions',
                                                    environment=environment, username=None, password=None)

    def new(self, transaction):
        self.logger.info('New Transaction - Asset Manager: %s - Transaction ID: %s', transaction.asset_manager_id,
                         transaction.transaction_id)
        url = '%s/transactions/%s' % (self.endpoint, transaction.asset_manager_id)
        response = self.session.post(url, json=transaction.to_interface())
        if response.ok:
            transaction = json_to_transaction(response.json())
            return transaction
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def create_many(self, transactions):
        if type(transactions) is not list:
            raise ValueError('Error - create_many takes in a list of transactions instead of single transaction')
        transactions_json = []        
        # check to ensure all transactions have the same asset_manager_id
        asset_manager_id = transactions[0].asset_manager_id
        for transaction in transactions:
            transactions_json.append(transaction.to_interface())
            if transaction.asset_manager_id != asset_manager_id:
                raise AttributeError('Check failed - Not all transactions have the same asset manager ID.')
        self.logger.info('Multple new Transactions - Asset Manager: %s', asset_manager_id)
        url = '%s/transactions/%s' % (self.endpoint, asset_manager_id)
        response = self.session.post(url, json=transactions_json)
        if response.ok:
            transactions = []
            for transaction_json in response.json():
                transactions.append(json_to_transaction(transaction_json))
            return transactions
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def amend(self, transaction):
        self.logger.info('Amend Transaction - Asset Manager: %s - Transaction ID: %s', transaction.asset_manager_id,
                         transaction.transaction_id)
        url = '%s/transactions/%s/%s' % (self.endpoint, transaction.asset_manager_id, transaction.transaction_id)
        response = self.session.put(url, json=transaction.to_interface())
        if response.ok:
            transaction = json_to_transaction(response.json())
            return transaction
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def partial(self, asset_manager_id, transaction_id, updates):
        self.logger.info('Partial Amend Transaction - Asset Manager: %s - Transaction ID: %s', asset_manager_id,
                         transaction_id)
        url = '%s/transactions/%s/%s' % (self.endpoint, asset_manager_id, transaction_id)
        response = self.session.patch(url, data=json.dumps(updates, default=json_handler), headers=self.json_header)
        if response.ok:
            transaction = json_to_transaction(response.json())
            return transaction
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def retrieve(self, asset_manager_id, transaction_id, version=None):
        self.logger.info('Retrieve Transaction - Asset Manager: %s - Transaction ID: %s', asset_manager_id,
                         transaction_id)
        url = '%s/transactions/%s/%s' % (self.endpoint, asset_manager_id, transaction_id)
        if version:
            url += '?version=%d' % int(version)
        response = self.session.get(url)
        if response.ok:
            return json_to_transaction(response.json())
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def transactions_by_asset_manager(self, asset_manager_id):
        self.logger.info('Retrieve Transactions by Asset Manager: %s', asset_manager_id)
        url = '%s/transactions/%s' % (self.endpoint, asset_manager_id)
        response = self.session.get(url)
        if response.ok:
            transactions = [json_to_transaction(json_transaction) for json_transaction in response.json()]
            self.logger.info('Returned %s Transactions.', len(transactions))
            return transactions
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def cancel(self, asset_manager_id, transaction_id):
        self.logger.info('Cancel Transaction - Asset Manager: %s - Transaction ID: %s', asset_manager_id,
                         transaction_id)
        url = '%s/transactions/%s/%s' % (self.endpoint, asset_manager_id, transaction_id)
        response = self.session.delete(url)
        if response.ok:
            self.logger.info('Successfully Cancelled - Asset Manager: %s - Transaction ID: %s.', asset_manager_id,
                             transaction_id)
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def search(self, asset_manager_id, transaction_ids=[], transaction_statuses=[],
               asset_book_ids=[], counterparty_book_ids=[], asset_ids=[], transaction_date_start=None,
               transaction_date_end=None, code_types=[], code_values=[], link_types=[], linked_transaction_ids=[],
               party_types=[], party_ids=[], reference_types=[], reference_values=[], client_ids=[],
               page_no=None, page_size=None):
        self.logger.info('Search Transactions - Asset Manager: %s', asset_manager_id)
        search_params = {}
        # Potentially roll this into a loop through args rather than explicitly named - depends on additional validation
        if transaction_ids:
            search_params['transaction_ids'] = ','.join(transaction_ids)
        if transaction_statuses:
            search_params['transaction_statuses'] = ','.join(transaction_statuses)
        if asset_book_ids:
            search_params['asset_book_ids'] = ','.join(asset_book_ids)
        if counterparty_book_ids:
            search_params['counterparty_book_ids'] = ','.join(counterparty_book_ids)
        if asset_ids:
            search_params['asset_ids'] = ','.join(asset_ids)
        if transaction_date_start:
            search_params['transaction_date_start'] = transaction_date_start
        if transaction_date_end:
            search_params['transaction_date_end'] = transaction_date_end
        if code_types:
            search_params['code_types'] = ','.join(code_types)
        if code_values:
            search_params['code_values'] = ','.join(code_values)
        if link_types:
            search_params['link_types'] = ','.join(link_types)
        if linked_transaction_ids:
            search_params['linked_transaction_ids'] = ','.join(linked_transaction_ids)
        if party_types:
            search_params['party_types'] = ','.join(party_types)
        if party_ids:
            search_params['party_ids'] = ','.join(party_ids)
        if reference_types:
            search_params['reference_types'] = ','.join(reference_types)
        if reference_values:
            search_params['reference_values'] = ','.join(reference_values)
        if client_ids:
            search_params['client_ids'] = ','.join(client_ids)
        if page_no is not None:
            search_params['page_no'] = page_no
        if page_size:
            search_params['page_size'] = page_size
        url = '%s/transactions/%s' % (self.endpoint, asset_manager_id)
        response = self.session.get(url, params=search_params)
        if response.ok:
            transactions = [json_to_transaction(json_transaction) for json_transaction in response.json()]
            self.logger.info('Returned %s Transactions.', len(transactions))
            return transactions
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def new_mtm_results(self, asset_manager_id, mtm_results):
        self.logger.info('Marking to market Positions - Asset Manager: %s', asset_manager_id)
        if not isinstance(mtm_results, list):
            mtm_results = [mtm_results]
        mtm_result_json = []
        for mtm_result in mtm_results:
            mtm_result_json.append(mtm_result.to_interface())
        url = '%s/mtm/%s' % (self.endpoint, asset_manager_id)
        response = self.session.post(url, json=mtm_result_json)
        if response.ok:
            mtm_results = []
            for mtm_result_json in response.json():
                mtm_results.append(json_to_mtm_result(mtm_result_json))
            return mtm_results
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def upsert_mtm_results(self, asset_manager_id, mtm_results):
        self.logger.info('Upserting Marking to market Positions - Asset Manager: %s', asset_manager_id)
        if not isinstance(mtm_results, list):
            mtm_results = [mtm_results]
        mtm_result_json = []
        for mtm_result in mtm_results:
            mtm_result_json.append(mtm_result.to_interface())
        url = '%s/mtm/%s' % (self.endpoint, asset_manager_id)
        response = self.session.post(url, json=mtm_result_json, params={'upsert': True})
        if response.ok:
            mtm_results = []
            for mtm_result_json in response.json():
                mtm_results.append(json_to_mtm_result(mtm_result_json))
            return mtm_results
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def amend_mtm_results(self, asset_manager_id, mtm_results):
        self.logger.info('Amending mtm Positions - Asset Manager: %s', asset_manager_id)
        if not isinstance(mtm_results, list):
            mtm_results = [mtm_results]
        mtm_result_json = []
        for mtm_result in mtm_results:
            mtm_result_json.append(mtm_result.to_interface())
        url = '%s/mtm/%s' % (self.endpoint, asset_manager_id)
        response = self.session.put(url, json=mtm_result_json)
        if response.ok:
            mtm_results = []
            for mtm_result_json in response.json():
                mtm_results.append(json_to_mtm_result(mtm_result_json))
            return mtm_results
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def retrieve_mtm_results(self, book_id, asset_manager_id, paramaters):
        """
        parameters is a dictionary of all the mtm result filter parameters
        """
        self.logger.info('Retrieving mtm Positions - Asset Manager: %s', asset_manager_id)
        url = '%s/mtm/%s' % (self.endpoint, asset_manager_id)       
        paramaters.update({'book_id': book_id})  
        response = self.session.get(url, params = paramaters)
        if response.ok:
            mtm_results = [json_to_mtm_result(json_mtm_result) for json_mtm_result in response.json()]
            self.logger.info('Returned %s mtm results.', len(mtm_results))
            return mtm_results
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def retrieve_aggregate_pnls(self, asset_manager_id, book_ids, business_date, currency):
        self.logger.info('Retrieving Aggregate PnL result - Asset Manager: %s - Book IDs: (%s) - Business Date: %s - Currency: %s' % \
                        (asset_manager_id, ', '.join(book_ids), business_date, currency))
        url = '%s/aggregate_pnls/%s' % (self.endpoint, asset_manager_id)
        search_params = {'business_date': business_date,
                         'book_ids': book_ids,
                         'currency': currency}
        response = self.session.get(url, params=search_params)
        if response.ok:
            return response.json()
        else:
            self.logger.error(response.text)
            response.raise_for_status()    

    def new_transaction_pnls(self, asset_manager_id, transaction_pnls):
        self.logger.info('Insert Transaction PnL results for - Asset Manager: %s', asset_manager_id)
        if not isinstance(transaction_pnls, list):
            transaction_pnls = [transaction_pnls]
        transaction_pnl_json = []
        for transaction_pnl in transaction_pnls:
            transaction_pnl_json.append(transaction_pnl.to_interface())
        url = '%s/transaction_pnls/%s' % (self.endpoint, asset_manager_id)
        response = self.session.post(url, json=transaction_pnl_json)
        if response.ok:
            transaction_pnls = []
            for transaction_pnl_json in response.json():
                transaction_pnls.append(json_to_transaction_pnl(transaction_pnl_json))
            self.logger.info('Created %s Transaction PnL records', len(transaction_pnls))
            return transaction_pnls
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def upsert_transaction_pnls(self, asset_manager_id, transaction_pnls):
        self.logger.info('Upsert Transaction PnL results for - Asset Manager: %s', asset_manager_id)
        if not isinstance(transaction_pnls, list):
            transaction_pnls = [transaction_pnls]
        transaction_pnl_json = []
        for transaction_pnl in transaction_pnls:
            transaction_pnl_json.append(transaction_pnl.to_interface())
        url = '%s/transaction_pnls/%s' % (self.endpoint, asset_manager_id)
        response = self.session.post(url, json=transaction_pnl_json, params={'upsert': True})
        if response.ok:
            transaction_pnls = []
            for transaction_pnl_json in response.json():
                transaction_pnls.append(json_to_transaction_pnl(transaction_pnl_json))
            self.logger.info('Upserted %s Transaction PnL records', len(transaction_pnls))
            return transaction_pnls
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def amend_transaction_pnls(self, asset_manager_id, transaction_pnls):
        self.logger.info('Amend Transaction PnL results for - Asset Manager: %s', asset_manager_id)
        if not isinstance(transaction_pnls, list):
            transaction_pnls = [transaction_pnls]
        transaction_pnl_json = []
        for transaction_pnl in transaction_pnls:
            transaction_pnl_json.append(transaction_pnl.to_interface())
        url = '%s/transaction_pnls/%s' % (self.endpoint, asset_manager_id)
        response = self.session.put(url, json=transaction_pnl_json)
        if response.ok:
            transaction_pnls = []
            for transaction_pnl_json in response.json():
                transaction_pnls.append(json_to_transaction_pnl(transaction_pnl_json))
            self.logger.info('Amended %s Transaction PnL records', len(transaction_pnls))
            return transaction_pnls
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def retrieve_transaction_pnls(self, book_ids, asset_manager_id, business_date, periods=None):
        self.logger.info('Retrieving Transaction PnL result - Asset Manager: %s - Book IDs: (%s) - Business Date: %s' % \
                        (asset_manager_id, ', '.join(book_ids), business_date))
        url = '%s/transaction_pnls/%s' % (self.endpoint, asset_manager_id)
        search_params = {'business_date': business_date,
                         'book_ids': book_ids}
        if periods:
            search_params['periods'] = periods
        response = self.session.get(url, params=search_params)
        if response.ok:
            transaction_pnls = [json_to_transaction_pnl(json_transaction_pnl) for json_transaction_pnl in response.json()]
            self.logger.info('Returned %s Transaction PnL results.', len(transaction_pnls))
            return transaction_pnls
        else:
            self.logger.error(response.text)
            response.raise_for_status()
    
    def new_position_pnls(self, asset_manager_id, position_pnls):
        self.logger.info('Insert Position PnL results for - Asset Manager: %s', asset_manager_id)
        if not isinstance(position_pnls, list):
            position_pnls = [position_pnls]
        position_pnl_json = []
        for position_pnl in position_pnls:
            position_pnl_json.append(position_pnl.to_interface())
        url = '%s/position_pnls/%s' % (self.endpoint, asset_manager_id)
        response = self.session.post(url, json=position_pnl_json)
        if response.ok:
            position_pnls = []
            for position_pnl_json in response.json():
                position_pnls.append(json_to_position_pnl(position_pnl_json))
            self.logger.info('Created %s Position PnL records', len(position_pnls))
            return position_pnls
        else:
            self.logger.error(response.text)
            response.raise_for_status() 

    def upsert_position_pnls(self, asset_manager_id, position_pnls):
        self.logger.info('Upsert Position PnL results for - Asset Manager: %s', asset_manager_id)
        if not isinstance(position_pnls, list):
            position_pnls = [position_pnls]
        position_pnl_json = []
        for position_pnl in position_pnls:
            position_pnl_json.append(position_pnl.to_interface())
        url = '%s/position_pnls/%s' % (self.endpoint, asset_manager_id)
        response = self.session.post(url, json=position_pnl_json, params={'upsert': True})
        if response.ok:
            position_pnls = []
            for position_pnl_json in response.json():
                position_pnls.append(json_to_position_pnl(position_pnl_json))
            self.logger.info('Upserted %s Position PnL records', len(position_pnls))
            return position_pnls
        else:
            self.logger.error(response.text)
            response.raise_for_status() 

    def amend_position_pnls(self, asset_manager_id, position_pnls):
        self.logger.info('Amend Position PnL results for - Asset Manager: %s', asset_manager_id)
        if not isinstance(position_pnls, list):
            position_pnls = [position_pnls]
        position_pnl_json = []
        for position_pnl in position_pnls:
            position_pnl_json.append(position_pnl.to_interface())
        url = '%s/position_pnls/%s' % (self.endpoint, asset_manager_id)
        response = self.session.put(url, json=position_pnl_json)
        if response.ok:
            position_pnls = []
            for position_pnl_json in response.json():
                position_pnls.append(json_to_position_pnl(position_pnl_json))
            self.logger.info('Amended %s Position PnL records', len(position_pnls))
            return position_pnls
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def retrieve_position_pnls(self, book_ids, asset_manager_id, business_date, periods=None):
        self.logger.info('Retrieving Position PnL result - Asset Manager: %s - Book IDs: (%s) - Business Date: %s' % \
                        (asset_manager_id, ', '.join(book_ids), business_date))
        url = '%s/position_pnls/%s' % (self.endpoint, asset_manager_id)
        search_params = {'business_date': business_date,
                         'book_ids': book_ids}
        if periods:
            search_params['periods'] = periods
        response = self.session.get(url, params=search_params)
        if response.ok:
            position_pnls = [json_to_position_pnl(json_position_pnl) for json_position_pnl in response.json()]
            self.logger.info('Retrieved %s Position PnL results.', len(position_pnls))
            return position_pnls
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def get_pnl_transactions(self, book_id, asset_manager_id, start_date, end_date):
        self.logger.info('Retrieving transactions for PnL - Asset Manager: %s - Book ID: %s'%(asset_manager_id, book_id))
        url = '%s/pnl_transactions/%s' % (self.endpoint, asset_manager_id)
        search_params = {'asset_manager_id': asset_manager_id,
                         'start_date': start_date,
                         'book_id': book_id,
                         'end_date': end_date}
        response = self.session.get(url, params=search_params)
        if response.ok:
            transactions = [json_to_transaction(json_transaction) for json_transaction in response.json()]
            self.logger.info('Returned %s Transactions.', len(transactions))
            return transactions
        else:
            self.logger.error(response.text)
            response.raise_for_status()        

    def position_search(self, asset_manager_id, book_ids=None, account_ids=None,
                        accounting_types=None, asset_ids=None,
                        position_date=None, include_cash=False, 
                        page_no=None, page_size=None):
        self.logger.info('Search Positions - Asset Manager: %s', asset_manager_id)
        search_params = {}
        # Potentially roll into a loop
        if book_ids:
            search_params['book_ids'] = ','.join(book_ids)
        if account_ids:
            search_params['account_ids'] = ','.join(account_ids)
        if accounting_types:
            search_params['accounting_types'] = ','.join(accounting_types)
        if asset_ids:
            search_params['asset_ids'] = ','.join(asset_ids)
        if position_date:
            search_params['position_date'] = position_date
        if include_cash:
            search_params['include_cash'] = include_cash
        if page_no is not None:
            search_params['page_no'] = page_no
        if page_size:
            search_params['page_size'] = page_size
        url = '%s/positions/%s' % (self.endpoint, asset_manager_id)
        response = self.session.get(url, params=search_params)
        if response.ok:
            positions = [json_to_position(json_position) for json_position in response.json()]
            self.logger.info('Returned %s Positions.', len(positions))
            return positions
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    # Should this method just be collapsed into positions_by_asset_manager?
    def positions_by_asset_manager_book(self, asset_manager_id, book_id):
        self.logger.info('Retrieve Positions by Asset Manager: %s and Book: %s', asset_manager_id, book_id)
        url = '%s/positions/%s/%s' % (self.endpoint, asset_manager_id, book_id)
        response = self.session.get(url)
        if response.ok:
            positions = [json_to_position(json_position) for json_position in response.json()]
            self.logger.info('Returned %s Positions.', len(positions))
            return positions
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def positions_by_asset_manager(self, asset_manager_id, book_ids=None):
        self.logger.info('Retrieve Positions by Asset Manager: %s', asset_manager_id)
        url = '%s/positions/%s' % (self.endpoint, asset_manager_id)
        params = {'book_ids': ','.join(book_ids)} if book_ids else {}
        response = self.session.get(url, params=params)
        if response.ok:
            positions = [json_to_position(json_position) for json_position in response.json()]
            self.logger.info('Returned %s Positions.', len(positions))
            return positions
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def allocate_transaction(self, asset_manager_id, transaction_id, allocation_type, allocation_dicts):
        """

        :param asset_manager_id:
        :param transaction_id:
        :param allocation_type:
        :param allocation_dicts:
        :return:
        """
        self.logger.info('Allocate Transaction - Asset Manager: %s - Transaction ID: %s', asset_manager_id,
                         transaction_id)
        url = '%s/allocations/%s/%s' % (self.endpoint, asset_manager_id, transaction_id)
        params = {'allocation_type': allocation_type}
        response = self.session.post(url, params=params, data=json.dumps(allocation_dicts, default=json_handler),
                                     headers=self.json_header)
        if response.ok:
            allocations = [json_to_transaction(json_allocation) for json_allocation in response.json()]
            allocation_ids = [allocation.transaction_id for allocation in allocations]
            self.logger.info('%s Allocations Created - Transactions: %s', len(allocations), allocation_ids)
            return allocations
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def retrieve_transaction_allocations(self, asset_manager_id, transaction_id):
        """

        :param asset_manager_id:
        :param transaction_id:
        :return:
        """
        self.logger.info('Retrieve Allocations - Asset Manager: %s - Transaction ID: %s', asset_manager_id,
                         transaction_id)
        url = '%s/allocations/%s/%s' % (self.endpoint, asset_manager_id, transaction_id)
        response = self.session.get(url)
        if response.ok:
            allocations = [json_to_transaction(json_allocation) for json_allocation in response.json()]
            self.logger.info('Returned %s Allocations.', len(allocations))
            return allocations
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def net_transactions(self, asset_manager_id, transaction_ids, netting_type='Net'):
        """

        :param asset_manager_id: The asset_manager_id of the netting set owner
        :param transaction_ids:  A list of transaction_ids to net
        :param netting_type:
        :return:
        """
        self.logger.info('Net Transactions - Asset Manager: %s - Transaction IDs: %s', asset_manager_id,
                         transaction_ids)
        url = '%s/netting/%s' % (self.endpoint, asset_manager_id)
        params = {'netting_type': netting_type}
        response = self.session.post(url, params=params, json=transaction_ids)
        if response.ok:
            net_transaction = json_to_transaction(response.json())
            self.logger.info('Net Created - Transaction: %s', net_transaction.transaction_id)
            return net_transaction
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def retrieve_netting_set(self, asset_manager_id, transaction_id):
        """
        Returns all the transaction_ids associated with a single netting set.  Pass in the ID for any transaction in
        the set.
        :param asset_manager_id:  The asset_manager_id for the netting set owner.
        :param transaction_id: A transaction_id of an entry within the netting set.
        :return:
        """
        self.logger.info('Retrieve Netting Set - Asset Manager: %s - Transaction ID: %s', asset_manager_id,
                         transaction_id)
        url = '%s/netting/%s/%s' % (self.endpoint, asset_manager_id, transaction_id)
        response = self.session.get(url)
        if response.ok:
            net_transaction_id, netting_set_json = next(iter(response.json().items()))
            netting_set = [json_to_transaction(net_transaction) for net_transaction in netting_set_json]
            self.logger.info('Returned %s Transactions in Netting Set.', len(netting_set))
            return net_transaction_id, netting_set
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def book_transfer(self, asset_manager_id, asset_id, source_book_id, target_book_id, wash_book_id, quantity, price,
                      currency):
        """
        A method for moving between books.  The convention is always *from* the source, *to* the target.

        Two transactions are booked -  one against each

        :param asset_manager_id: The owning asset manager id.
        :param asset_id: The asset being transferred.
        :param source_book_id: The book id of the source.
        :param target_book_id:  The book id of the target.
        :param wash_book_id:  The book id of the wash book which will be the counterparty for the two sides.
        :param quantity: The quantity to transfer.
        :param price:  The price at which to make the transfer.  Typically T-1 EOD price or current market price.
        :param currency: The currency for the transfer price.
        :return:
        """
        url = '%s/book_transfer/%s' % (self.endpoint, asset_manager_id)
        body = {'asset_id': asset_id, 'source_book_id': source_book_id, 'target_book_id': target_book_id,
                'wash_book_id': wash_book_id, 'quantity': quantity, 'price': price, 'currency': currency}
        response = self.session.post(url, data=json.dumps(body, default=json_handler), headers=self.json_header)
        if response.ok:
            deliver_json, receive_json = response.json()
            return json_to_transaction(deliver_json), json_to_transaction(receive_json),
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def depot_transfer(self, asset_manager_id, asset_id, source_account_id, target_account_id, quantity):
        raise NotImplementedError("""This is not yet supported.  The concept is similar to a
                                  book transfer, except it requires an external message to a
                                  custodian to instruct them to move the stock to a
                                   different depot account.""")

    def clear(self, asset_manager_id, book_ids=None):
        """ This method deletes all the data for an asset_manager_id
            and option book_ids.
            It should be used with extreme caution.  In production it
            is almost always better to Inactivate rather than delete. """
        self.logger.info('Clear Transactions & Positions - Asset Manager: %s', asset_manager_id)
        url = '%s/clear/%s' % (self.endpoint, asset_manager_id)
        params = {'asset_manager_ids': ','.join(book_ids)} if book_ids else {}
        response = self.session.delete(url, params=params)
        if response.ok:
            tran_count = response.json().get('transaction_count', 'Unknown')
            self.logger.info('Deleted %s Transactions.', tran_count)
            pos_count = response.json().get('position_count', 'Unknown')
            self.logger.info('Deleted %s Positions.', pos_count)
            return response.json()
        else:
            self.logger.error(response.text)
            response.raise_for_status()


from __future__ import absolute_import, division, print_function, unicode_literals

from amaascore.corporate_actions.corporate_action import CorporateAction


class Notification(CorporateAction):

    def __init__(self, asset_manager_id, corporate_action_id, record_date, corporate_action_status='Open',
                 asset_id=None, party_id=None, declared_date=None, settlement_date=None, elective=False, message=None,
                 description='', references=None, *args, **kwargs):
        super(Notification, self).__init__(asset_manager_id=asset_manager_id, corporate_action_id=corporate_action_id,
                                           record_date=record_date, corporate_action_status=corporate_action_status,
                                           asset_id=asset_id, party_id=party_id, declared_date=declared_date,
                                           settlement_date=settlement_date, elective=elective, message=message,
                                           description=description, references=references, *args, **kwargs)



from decimal import Decimal
from dateutil.parser import parse
import pytz
from amaascore.core.amaas_model import AMaaSModel


class MTMResult(AMaaSModel):
    def __init__(self, asset_manager_id, book_id, business_date, asset_id,
                 mtm_timestamp, mtm_value=None, mtm_status='Active', message=None,
                 *args, **kwargs):
        self.asset_manager_id = asset_manager_id
        self.asset_id = asset_id
        self.book_id = book_id
        self.mtm_value = mtm_value
        self.business_date = business_date
        self.mtm_timestamp = mtm_timestamp
        self.message = message
        self.mtm_status = mtm_status

        super(MTMResult, self).__init__(*args, **kwargs)

    @property
    def mtm_timestamp(self):
        return self._mtm_timestamp

    @mtm_timestamp.setter
    def mtm_timestamp(self, value):
        if value:
            mtm_timestamp = parse(value) if isinstance(value, str) else value
            self._mtm_timestamp = mtm_timestamp.replace(tzinfo=pytz.UTC) \
                if not mtm_timestamp.tzinfo else mtm_timestamp

    @property
    def mtm_value(self):
        return self._mtm_value

    @mtm_value.setter
    def mtm_value(self, val):
        if not isinstance(val, Decimal) and val is not None:
            self._mtm_value = Decimal(val)
        else:
            self._mtm_value = val


# POTENTIALLY SUPPORT MULTIPLE ENVIRONMENTS
from __future__ import absolute_import, division, print_function, unicode_literals
from collections import namedtuple

ENVIRONMENT = 'production'
ENDPOINTS = {
    'asset_managers': 'assetmanager',
    'assets': 'asset',
    'books': 'book',
    'corporate_actions': 'corporateaction',
    'fundamentals': 'fundamental',
    'market_data': 'marketdata',
    'monitor': 'monitor',
    'parties': 'party',
    'transactions': 'transaction'
}

EnvironmentConfig = namedtuple('EnvironmentConfig',
                               'environment base_url api_version cognito_region cognito_pool cognito_client_id')
CONFIGURATIONS = {'production': EnvironmentConfig(environment='production',
                                                  base_url='https://api.amaas.com',
                                                  api_version='sg1.0',
                                                  cognito_region='ap-southeast-1',
                                                  cognito_pool='ap-southeast-1_0LilJdUR3',
                                                  cognito_client_id='7b7kt883veb7rr2toolj1ukp6j'),
                  'staging': EnvironmentConfig(environment='staging',
                                               base_url='https://api-staging.amaas.com',
                                               api_version='v1.0',
                                               cognito_region='ap-southeast-1',
                                               cognito_pool='ap-southeast-1_De6j7TWIB',
                                               cognito_client_id='2qk35mhjjpk165vssuqhqoi1lk'),
                  'staging-pro': EnvironmentConfig(environment='staging-pro',
                                                   base_url='https://api-staging-pro.amaas.com',
                                                   api_version='v1.0',
                                                   cognito_region='ap-southeast-1',
                                                   cognito_pool='ap-southeast-1_De6j7TWIB',
                                                   cognito_client_id='2qk35mhjjpk165vssuqhqoi1lk'),
                  'dev': EnvironmentConfig(environment='dev',
                                           base_url='https://api-dev.amaas.com',
                                           api_version='v1.0',
                                           cognito_region='ap-southeast-1',
                                           cognito_pool='ap-southeast-1_De6j7TWIB',
                                           cognito_client_id='2qk35mhjjpk165vssuqhqoi1lk'),
                  'local': EnvironmentConfig(environment='local',
                                             base_url='http://localhost:8000',
                                             api_version='',
                                             cognito_region='ap-southeast-1',
                                             cognito_pool='ap-southeast-1_De6j7TWIB',
                                             cognito_client_id='2qk35mhjjpk165vssuqhqoi1lk')}




from __future__ import absolute_import, division, print_function, unicode_literals

from amaascore.parties.organisation import Organisation


class GovernmentAgency(Organisation):

    def __init__(self, asset_manager_id, party_id, description='', party_status='Active', 
                 display_name='', legal_name='', url='',
                 addresses=None, comments=None, emails=None, links=None, references=None,
                 *args, **kwargs):
        super(GovernmentAgency, self).__init__(asset_manager_id=asset_manager_id, party_id=party_id,
                                               description=description, party_status=party_status,
                                               display_name=display_name, legal_name=legal_name, url=url,
                                               addresses=addresses, comments=comments, emails=emails,
                                               links=links, references=references, *args, **kwargs)



from __future__ import absolute_import, division, print_function, unicode_literals

from amaascore.parties.company import Company


class AssetManager(Company):
    """
    This represents a Company engaged in Asset Management activity.
    """

    def __init__(self, asset_manager_id, party_id, base_currency=None, display_name='',
                 legal_name='', url='', description='', party_status='Active',
                 year_of_incorporation=None, contact_number=None, license_type=None,
                 license_number=None, assets_under_management=None, registration_number=None,
                 addresses=None, emails=None, links=None, references=None,
                 *args, **kwargs):
        self.license_type = license_type
        self.license_number = license_number
        self.assets_under_management = assets_under_management
        self.registration_number = registration_number
        super(AssetManager, self).__init__(asset_manager_id=asset_manager_id, party_id=party_id,
                                           base_currency=base_currency, display_name=display_name,
                                           legal_name=legal_name, url=url,
                                           description=description, party_status=party_status,
                                           year_of_incorporation=year_of_incorporation,
                                           contact_number=contact_number,
                                           addresses=addresses, emails=emails,
                                           links=links, references=references, *args, **kwargs)



from datetime import datetime, date
from dateutil import parser
from dateutil.parser import parse

from amaascore.assets.asset import Asset
from amaascore.assets.equity import Equity

class Warrant(Equity):

    @staticmethod
    def mandatory_attributes():
        return set()

    @staticmethod
    def additional_attributes():
        return set()

    def __init__(self, asset_manager_id, asset_id, client_id, asset_issuer_id=None, asset_status='Active',
                 description='', country_id=None, venue_id=None, currency=None, issue_date=date.min,
                 share_class=None, additional=None, display_name='', roll_price=False,
                 underlying_price=None, strike_price=None, cv_ratio=None,
                 exercise_type=None, expiration_date=None, first_exercise_date=None,
                 last_trade_date=None, issue_price=None, issued_quant=None,
                 outstanding=None, issuer=None, settlement_type=None, warrant_type=None, 
                 links=None, references=None, *args, **kwargs):
        super(Warrant, self).__init__(asset_manager_id=asset_manager_id, asset_id=asset_id, client_id=client_id, asset_issuer_id=asset_issuer_id, asset_status=asset_status,
                                     description=description, country_id=country_id, venue_id=venue_id, currency=currency, issue_date=issue_date,
                                     share_class=share_class, additional=additional, display_name=display_name, roll_price=roll_price,
                                     links=links, references=references, *args, **kwargs)

        #self.asset_class = 'Warrant'

        self.underlying_price = underlying_price
        self.strike_price = strike_price
        self.cv_ratio = cv_ratio
        self.exercise_type = exercise_type
        self.expiration_date = expiration_date
        self.first_exercise_date = first_exercise_date
        self.last_trade_date = last_trade_date
        self.issue_price = issue_price
        self.issued_quant = issued_quant
        self.outstanding = outstanding
        self.issuer = issuer
        self.settlement_type = settlement_type
        self.warrant_type = warrant_type

    @property
    def underlying_price(self):
        return self._underlying_price

    @underlying_price.setter
    def underlying_price(self, underlying_price):
        self._underlying_price = underlying_price
    
    @property
    def strike_price(self):
        return self._strike_price

    @strike_price.setter
    def strike_price(self, strike_price):
        self._strike_price = strike_price

    @property
    def cv_ratio(self):
        return self._cv_ratio

    @cv_ratio.setter
    def cv_ratio(self, cv_ratio):
        self._cv_ratio = cv_ratio

    @property
    def exercise_type(self):
        return self._exercise_type

    @exercise_type.setter
    def exercise_type(self, exercise_type):
        if exercise_type is None:
            self._exercise_type = None
        else:
            if exercise_type not in ['European', 'American']:
                raise ValueError("Invalid exercise_type European/American: %s" % exercise_type)
            self._exercise_type = exercise_type

    @property
    def expiration_date(self):
        return self._expiration_date

    @expiration_date.setter
    def expiration_date(self, expiration_date):
        self._expiration_date = expiration_date

    @property
    def first_exercise_date(self):
        return self._first_exercise_date

    @first_exercise_date.setter
    def first_exercise_date(self, first_exercise_date):
        self._first_exercise_date = first_exercise_date
    
    @property
    def last_trade_date(self):
        return self._last_trade_date

    @last_trade_date.setter
    def last_trade_date(self, last_trade_date):
        self._last_trade_date = last_trade_date

    @property
    def issue_price(self):
        return self._issue_price

    @issue_price.setter
    def issue_price(self, issue_price):
        self._issue_price = issue_price

    @property
    def issued_quant(self):
        return self._issued_quant

    @issued_quant.setter
    def issued_quant(self, issued_quant):
        self._issued_quant = issued_quant
    
    @property
    def outstanding(self):
        return self._outstanding

    @outstanding.setter
    def outstanding(self, outstanding):
        self._outstanding = outstanding

    @property
    def issuer(self):
        return self._issuer

    @issuer.setter
    def issuer(self, issuer):
        self._issuer = issuer
        
    @property
    def settlement_type(self):
        return self._settlement_type

    @settlement_type.setter
    def settlement_type(self, settlement_type):
        if settlement_type is None:
            self._settlement_type = None
        else:
            if settlement_type not in ['Cash', 'Physical']:
                raise ValueError("Invalid settlement_type Cash/Physical" % settlement_type)
            self._settlement_type = settlement_type

    @property
    def warrant_type(self):
        return self._warrant_type

    @warrant_type.setter
    def warrant_type(self, warrant_type):
        if warrant_type is None:
            self._warrant_type = None
        else:
            if warrant_type not in ["Covered", "Traditional"]:
                raise ValueError("Invalid warrant_type Covered/Traditional: %s" % warrant_type)
            self._warrant_type = warrant_type




from __future__ import absolute_import, division, print_function, unicode_literals

from dateutil.parser import parse
import sys

from amaascore.assets.asset import Asset

# This extremely ugly hack is due to the whole Python 2 vs 3 debacle.
type_check = str if sys.version_info >= (3, 0, 0) else (str, unicode)


class ForeignExchangeBase(Asset):
    """ This class should never be instantiated """

    def __init__(self, asset_id, asset_status, description, country_codes, display_name=None, currency=None,
                 asset_manager_id=0, *args, **kwargs):
        self.asset_class = 'ForeignExchange'
        self.country_codes = country_codes
        super(ForeignExchangeBase, self).__init__(asset_manager_id=asset_manager_id, asset_id=asset_id, fungible=True,
                                                  display_name=display_name or asset_id, roll_price=False,
                                                  asset_status=asset_status, description=description, currency=currency,
                                                  *args, **kwargs)

    def base_currency(self):
        return self.asset_id[0:3]

    def counter_currency(self):
        return self.asset_id[3:6]

    def get_country_codes(self):
        return self.country_codes

    def get_currencies(self):
        return [self.base_currency(), self.counter_currency()]

    @property
    def country_codes(self):
        return self._country_codes

    @country_codes.setter
    def country_codes(self, country_codes):
        """

        :param country_codes:
        :return:
        """
        self._country_codes = country_codes

class ForeignExchange(ForeignExchangeBase):
    """
    The underlying FX pair used in an FX spot/forward asset
    """
    def __init__(self, asset_id, asset_manager_id=0, asset_status='Active', country_codes=[], major=False,
                 description='', *args, **kwargs):
        self.major = major
        super(ForeignExchange, self).__init__(asset_manager_id=asset_manager_id, asset_id=asset_id,
                                              asset_status=asset_status, description=description,
                                              country_codes=country_codes,
                                              *args, **kwargs)

    @property
    def major(self):
        return self._major

    @major.setter
    def major(self, major):
        """

        :param major:
        :return:
        """
        if major:
            self._major = major
        else:
            self._major = False


class ForeignExchangeSpot(ForeignExchangeBase):
    """
    Spot FX (Settles as soon as possible).
    """

    def __init__(self, asset_manager_id, asset_id, asset_status='Active', description='', country_codes=[],
                 underlying=None, settlement_date=None, currency=None,
                 display_name=None, *args, **kwargs):
        self.underlying = underlying
        self.settlement_date = settlement_date
        super(ForeignExchangeSpot, self).__init__(asset_manager_id=asset_manager_id, asset_id=asset_id,
                                                  asset_status=asset_status, currency=currency,
                                                  country_codes=country_codes, display_name=display_name,
                                                  description=description, *args, **kwargs)

    @property
    def settlement_date(self):
        return self._settlement_date

    @settlement_date.setter
    def settlement_date(self, settlement_date):
        self._settlement_date = parse(settlement_date).date() if isinstance(settlement_date, type_check) \
            else settlement_date

    def base_currency(self):
        return self.underlying[0:3] if self.underlying else None

    def counter_currency(self):
        return self.underlying[3:6] if self.underlying else None


class ForeignExchangeForward(ForeignExchangeSpot):
    """
    A forward-dated FX.  If there is a fixing_date, it is an NDF.
    """

    def __init__(self, asset_manager_id, asset_id, asset_status='Active', description='', country_codes=[],
                 forward_rate=None, underlying=None, settlement_date=None, fixing_date=None, currency=None,
                 display_name=None, *args, **kwargs):
        self.forward_rate = forward_rate
        self.fixing_date = fixing_date
        super(ForeignExchangeForward, self).__init__(asset_manager_id=asset_manager_id, asset_id=asset_id,
                                                     asset_status=asset_status, currency=currency,
                                                     underlying=underlying, settlement_date=settlement_date,
                                                     country_codes=country_codes, display_name=display_name,
                                                     description=description, *args, **kwargs)

    @property
    def fixing_date(self):
        return self._fixing_date

    @fixing_date.setter
    def fixing_date(self, fixing_date):
        self._fixing_date = parse(fixing_date).date() if isinstance(fixing_date, type_check) \
            else fixing_date
        


from __future__ import absolute_import, division, print_function, unicode_literals

from amaascore.parties.company import Company


class Fund(Company):

    def __init__(self, asset_manager_id, party_id, base_currency, description='', party_status='Active',
                 display_name='', legal_name='', url='', 
                 addresses=None, comments=None, emails=None, links=None, references=None,
                 *args, **kwargs):
        super(Fund, self).__init__(asset_manager_id=asset_manager_id, party_id=party_id, base_currency=base_currency,
                                   description=description, party_status=party_status,
                                   display_name=display_name, legal_name=legal_name, url=url, 
                                   addresses=addresses, comments=comments, emails=emails,
                                   links=links, references=references,
                                   *args, **kwargs)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

CAR_COLOR_RECORD = {'White', 'Silver', 'Black', 'Grey', 'Blue', 'Red', 'Brown', 'Green'}
CAR_MAKE_RECORD = {'McLaren', 'Koenigsegg', 'Bugatti', 'Pagani', 'Aston Martin', 'Mercedes', 'Jaguar', 'Porsche',
            'Ferrari', 'Lamborghini', 'Rolls Royce', 'Bentley', 'Maybach', 'Audi', 'Maserati', 'Lexus', 'BMW', 'Cadillac'}
CAR_STYLE = {'R': 'Roadster', 'CV': 'Convertible', 'CP': 'Coupe', 'S': 'Sedan', 'SUV': 'Sport utility vehicle'}
CAR_GENRE = {'M': 'Modern', 'C': 'Classic', 'A': 'Antique', 'E': 'Exotic'}
CAR_RARITY = {'super rare', 'very rare', 'rare', 'fairly rare', None}
CAR_CONDITION = {'DC': 'Driving Condition', 'SS': 'Street Show Condition', 'SC': 'Show Car', 'CC': 'Concours Condition'}
CAR_PETROL_GRADE = {87: 'regular', 88: 'midgrade', 89: 'midgrade',
                    90: 'midgrade', 91: 'premium', 92: 'premium', 93: 'premium', 94: 'premium'}

CRYPTOCURRENCY_PROOF_TYPES = {'Proof of Work', 'Proof of Stake'}

PRICING_METHOD = {'Derived', 'Market'}

PRIVATE_INVESTMENT_CATEGORY = {'Private Equity', 'Mutual Funds', 'Hedge Funds', 'Fund of Funds', 'StartUp', 'Private Company', 'Others'}
PRIVATE_INVESTMENT_SUBCATEGORY = {'Mutual Funds': {'Money Market Funds', 'Bond Funds', 'Balanced Funds', 'Equity Funds',
                                                   'Speciality Funds'},
                                  'Private Equity': {'Leverage Buyout Funds', 'Growth Equity Funds', 'Venture Capital Funds',
                                                     'Real Estate Investment Funds', 'Mezzanine Funds', 'Distresses Debt Funds',
                                                     'Special Situation Funds'},
                                  'Hedge Funds': {'Market Neutral Funds', 'Equity Long Funds', 'Equity Short Funds', 'Event-Driven Funds'
                                                  'Macro Funds', 'Currency Funds', 'Equity Funds', 'Commodity', 'Credit', 'Index Arbitrage', 
                                                  'Quant Funds'},
                                  'Fund of Funds': {'Market Neutral Funds', 'Equity Long Funds', 'Equity Short Funds', 'Event-Driven Funds'
                                                    'Macro Funds', 'Currency Funds', 'Equity Funds', 'Commodity', 'Credit', 'Index Arbitrage', 
                                                    'Quant Funds'},
                                  'StartUp': {'Lifestyle Startup', 'Small Business', 'Silicon Valley Type Startup', 'Startup to be Quickly Sold'
                                              'Large Company Startup', 'Social Startups'}}
PRIVATE_INVESTMENT_SHARE_TYPE = {'Ordinary Shares', 'Defered Ordinary Shares', 'Non-voting Ordinary Shares', 'Redeemable Shares',
                                 'Redeemable Participating Shares', 'Redeemable Preference Shares', 'Preference Shares',
                                 'Cumulative Preference Shares', 'Participating Prefered Shares', 'Prefered Convertible Shares',
                                 'Prefered Convertible Shares', 'Convertible Note', 'Loan'}

WINE_CLASSIFICATIONS = {'1er Cru', '1er Cru Classé', '1er Grand Cru', '2ème Cru Classé', '2ème Vin',
                        '4ème Cru Classé', 'Cru Classé', 'Grand Cru', 'Non Classé'}
WINE_PACKING_TYPE = {'OCB06', 'OCB12', 'OWC06', 'OWC01', 'OWC02', 'OWC03', 'OWC06', 'OWC12'}





from __future__ import absolute_import, division, print_function, unicode_literals

from datetime import date, datetime
from dateutil.parser import parse
from decimal import Decimal
import sys
import pytz
import json

from amaascore.core.amaas_model import AMaaSModel

# This extremely ugly hack is due to the whole Python 2 vs 3 debacle.
type_check = str if sys.version_info >= (3, 0, 0) else (str, unicode)


class Curve(AMaaSModel):

    def __init__(self, asset_manager_id, asset_id, fixing_type, business_date, curve_timestamp, curve_rates, client_id, curve_type, additional, active=True,
                  *args, **kwargs):
        """

        :param asset_manager_id: The asset_manager_id of the asset for which we have the price.  Can be 0 for shared.
        :param asset_id: The FX pair for which this is the rate
        :param business_date: The business date for which this rate is used
        :param curve_timestamp: The timestamp for which this curve rate is valid
        :param curve_rates: The rates data points on curve, stored in json format
        :param curve_type: The type of curve, e.g. fx forward curve, interest rate curve etc.
        :param fixing_type: The fixing type of curve, EOD fixing or specific fixing like JPY forward curve at Tokyo closing
        :param active: The boolean status of the rate.  This is to allow rate corrections.
        :param additiona: any additional information about the curve
        """
        self.asset_manager_id = asset_manager_id
        self.asset_id = asset_id
        self.business_date = business_date
        self.curve_timestamp = curve_timestamp
        self.curve_type = curve_type
        self.curve_rates = curve_rates
        self.fixing_type = fixing_type
        self.active = active
        self.client_id = client_id
        self.additional = additional
        super(Curve, self).__init__(*args, **kwargs)
        

    def tenor_dates(self):
        additional = json.loads(self._additional)
        td = additional.get('tenor_dates')
        if td is None:
            raise ValueError("Error in retrieving tenor dates from database for %s on %s" %(self.asset_id, self.business_date))
        tenor_dates = {}     
        for tenor, date_str in td.items():
            date_elements = date_str.split('-')
            tenor_dates[tenor] = date(int(date_elements[0]), int(date_elements[1]), int(date_elements[2]))
        return tenor_dates


    @property
    def curve_rates(self):
        return self._curve_rates

    @curve_rates.setter
    def curve_rates(self, curve_rates):
        """
        :param curve_rates:
        :return:
        """
        if curve_rates is not None:
            if isinstance(curve_rates, str):
                self._curve_rates = curve_rates
            else:
                try:
                    self._curve_rates = json.dumps(curve_rates)
                except TypeError:
                    raise TypeError('Passed in curve rates object are not json serializable.Please check the format')

    @property
    def additional(self):
        return self._additional

    @additional.setter
    def additional(self, additional):
        if additional is not None:
            try:
                self._additional = json.dumps(additional)
            except TypeError:
                raise TypeError('The passed in "Additional" field is not json serializable.')
    @property
    def curve_timestamp(self):
        return self._curve_timestamp

    @curve_timestamp.setter
    def curve_timestamp(self, curve_timestamp):
        """
        Force the timestamp to be a datetime
        :param curve_timestamp:
        :return:
        """
        if curve_timestamp is not None:
            curve_timestamp = str(curve_timestamp)
            if isinstance(curve_timestamp, str):
                curve_timestamp = parse(curve_timestamp).replace(tzinfo=pytz.utc)
            if type(curve_timestamp) == date:
                curve_timestamp = datetime.combine(curve_timestamp, datetime.min.time()).replace(tzinfo=pytz.utc)
            if not curve_timestamp.tzinfo:
                raise ValueError('Cannot set a curve without a timezone')            
            if curve_timestamp > datetime.utcnow().replace(tzinfo=pytz.utc):
                raise ValueError('Cannot set a curve in the future: %s' % curve_timestamp.isoformat())
            self._curve_timestamp = curve_timestamp

    @property
    def business_date(self):
        return self._business_date

    @business_date.setter
    def business_date(self, business_date):
        """
        Force the business_date to always be a date
        :param business_date:
        :return:
        """
        if business_date is not None:
            if isinstance(business_date, str):
                self._business_date = parse(business_date).date()
            else:
                self._business_date= business_date   


from __future__ import absolute_import, division, print_function, unicode_literals

from amaascore.assets.currency import CurrencyBase
from amaascore.assets.enums import CRYPTOCURRENCY_PROOF_TYPES


class Cryptocurrency(CurrencyBase):

    def __init__(self, asset_id, asset_manager_id=0, asset_status='Active', proof_type='Proof of Work',
                 max_supply=None, display_name='', minor_unit_places=8, description='', country_id=None,
                 *args, **kwargs):
        self.proof_type = proof_type
        self.max_supply = max_supply
        super(Cryptocurrency, self).__init__(asset_manager_id=asset_manager_id, asset_id=asset_id,
                                             display_name=display_name,
                                             minor_unit_places=minor_unit_places,
                                             asset_status=asset_status, description=description,
                                             country_id=country_id, *args, **kwargs)

    @property
    def max_supply(self):
        if hasattr(self, '_max_supply'):
            return self._max_supply

    @max_supply.setter
    def max_supply(self, max_supply):
        """
        :param max_supply:
        :return:
        """
        self._max_supply = max_supply

    @property
    def proof_type(self):
        return self._proof_type

    @proof_type.setter
    def proof_type(self, proof_type):
        """
        :param proof_type:
        :return:
        """
        if proof_type in CRYPTOCURRENCY_PROOF_TYPES:
            self._proof_type = proof_type
        else:
            raise ValueError("Invalid input for proof type: %s" % proof_type)



from __future__ import absolute_import, division, print_function, unicode_literals

from amaascore.asset_managers.asset_manager import AssetManager
from amaascore.asset_managers.domain import Domain
from amaascore.asset_managers.relationship import Relationship

def json_to_asset_manager(json_asset_manager):
    asset_manager = AssetManager(**json_asset_manager)
    return asset_manager


def json_to_relationship(json_relationship):
    relationship = Relationship(**json_relationship)
    return relationship

def json_to_domain(json_domain):
    domain = Domain(**json_domain)
    return domain



from __future__ import absolute_import, division, print_function, unicode_literals

from decimal import Decimal

from amaascore.assets.future import Future
from amaascore.assets.option_mixin import OptionMixin


class FutureOption(Future, OptionMixin):

    def __init__(self, asset_manager_id, asset_id, option_type, option_style, strike, underlying_asset_id,
                 settlement_type, contract_size, point_value, tick_size, quote_unit=None, currency=None,
                 asset_issuer_id=None, asset_status='Active', issue_date=None, expiry_date=None, display_name='',
                 description='', country_id=None, venue_id=None, links=None, references=None, *args, **kwargs):
        self.option_type = option_type
        self.option_style = option_style
        self.strike = strike
        super(FutureOption, self).__init__(asset_manager_id=asset_manager_id, asset_id=asset_id,
                                           asset_issuer_id=asset_issuer_id, asset_status=asset_status,
                                           display_name=display_name, currency=currency,
                                           description=description, country_id=country_id, venue_id=venue_id,
                                           links=links, references=references, issue_date=issue_date,
                                           expiry_date=expiry_date, settlement_type=settlement_type,
                                           contract_size=contract_size, point_value=point_value,
                                           tick_size=tick_size, quote_unit=quote_unit,
                                           underlying_asset_id=underlying_asset_id,
                                           *args, **kwargs)



from __future__ import absolute_import, division, print_function, unicode_literals

import json
import logging

from amaascore.assets.utils import json_to_asset
from amaascore.config import ENVIRONMENT
from amaascore.core.interface import Interface
from amaascore.core.amaas_model import json_handler


class AssetsInterface(Interface):

    def __init__(self, environment=ENVIRONMENT, endpoint=None, logger=None, username=None, password=None):
        self.logger = logger or logging.getLogger(__name__)
        super(AssetsInterface, self).__init__(endpoint=endpoint, endpoint_type='assets',
                                              environment=environment, username=username, password=password)

    def new(self, asset):
        self.logger.info('New Asset - Asset Manager: %s - Asset ID: %s', asset.asset_manager_id, asset.asset_id)
        url = '%s/assets/%s' % (self.endpoint, asset.asset_manager_id)
        response = self.session.post(url, json=asset.to_interface())
        if response.ok:
            self.logger.info('Successfully Created Asset - Asset Manager: %s - Asset ID: %s', asset.asset_manager_id,
                             asset.asset_id)
            asset = json_to_asset(response.json())
            return asset
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def create_many(self, assets):
        if not assets or not isinstance(assets, list):
            raise ValueError('Invalid argument. Argument must be a non-empty list.')

        self.logger.info('New Assets - Asset Manager: %s', assets[0].asset_manager_id)
        url = '%s/assets/%s' % (self.endpoint, assets[0].asset_manager_id)
        json_body = [asset.to_interface() for asset in assets]
        response = self.session.post(url, json=json_body)
        if response.ok:
            self.logger.info('Successfully Created Assets - Asset Manager: %s', assets[0].asset_manager_id)
            assets = [asset for asset in response.json()]
            return assets
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def upsert(self, asset):
        ''' upsert only support upserting one asset at a time'''
        self.logger.info('Upsert Asset - Asset Manager: %s - Asset ID: %s', asset.asset_manager_id, asset.asset_id)
        url = '%s/assets/%s' % (self.endpoint, asset.asset_manager_id)
        response = self.session.post(url, json=asset.to_interface(), params={'upsert': True})
        if response.ok:
            self.logger.info('Successfully Upserted Asset - Asset Manager: %s - Asset ID: %s', asset.asset_manager_id,
                             asset.asset_id)
            asset = json_to_asset(response.json())
            return asset
        else:
            self.logger.error(response.text)
            response.raise_for_status()


    def amend(self, asset):
        self.logger.info('Amend Asset - Asset Manager: %s - Asset ID: %s', asset.asset_manager_id, asset.asset_id)
        url = '%s/assets/%s/%s' % (self.endpoint, asset.asset_manager_id, asset.asset_id)
        response = self.session.put(url, json=asset.to_interface())
        if response.ok:
            self.logger.info('Successfully Amended Asset - Asset Manager: %s - Asset ID: %s', asset.asset_manager_id,
                             asset.asset_id)
            asset = json_to_asset(response.json())
            return asset
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def partial(self, asset_manager_id, asset_id, updates):
        self.logger.info('Partial Amend Asset - Asset Manager: %s - Asset ID: %s', asset_manager_id,
                         asset_id)
        url = '%s/assets/%s/%s' % (self.endpoint, asset_manager_id, asset_id)
        # Setting handler ourselves so we can be sure Decimals work
        response = self.session.patch(url, data=json.dumps(updates, default=json_handler), headers=self.json_header)
        if response.ok:
            asset = json_to_asset(response.json())
            return asset
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def retrieve(self, asset_manager_id, asset_id, version=None):
        self.logger.info('Retrieve Asset - Asset Manager: %s - Asset ID: %s', asset_manager_id, asset_id)
        url = '%s/assets/%s/%s' % (self.endpoint, asset_manager_id, asset_id)
        if version:
            url += '?version=%d' % int(version)
        response = self.session.get(url)
        if response.ok:
            self.logger.info('Successfully Retrieved Asset - Asset Manager: %s - Asset ID: %s', asset_manager_id,
                             asset_id)
            return json_to_asset(response.json())
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def deactivate(self, asset_manager_id, asset_id):
        self.logger.info('Deactivate Asset - Asset Manager: %s - Asset ID: %s', asset_manager_id, asset_id)
        url = '%s/assets/%s/%s' % (self.endpoint, asset_manager_id, asset_id)
        json = {'asset_status': 'Inactive'}
        response = self.session.patch(url, json=json)
        if response.ok:
            self.logger.info('Successfully Deactivated Asset - Asset Manager: %s - Asset ID: %s', asset_manager_id,
                             asset_id)
            return json_to_asset(response.json())
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def search(self, asset_manager_id, asset_ids=None, asset_classes=None, asset_types=None,
               page_no=None, page_size=None):
        self.logger.info('Search for Assets - Asset Manager: %s', asset_manager_id)
        search_params = {}
        # Potentially roll this into a loop through args rather than explicitly named - depends on additional validation
        if asset_ids:
            search_params['asset_ids'] = ','.join(asset_ids)
        if asset_classes:
            search_params['asset_classes'] = ','.join(asset_classes)
        if asset_types:
            search_params['asset_types'] = ','.join(asset_types)
        if page_no is not None:
            search_params['page_no'] = page_no
        if page_size:
            search_params['page_size'] = page_size
        url = '%s/assets/%s' % (self.endpoint, asset_manager_id)
        response = self.session.get(url, params=search_params)
        if response.ok:
            assets = [json_to_asset(json_asset) for json_asset in response.json()]
            self.logger.info('Returned %s Assets.', len(assets))
            return assets
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def fields_search(self, asset_manager_id, asset_ids=None,
                      asset_classes=None, asset_types=None,
                      fields=None, page_no=None, page_size=None):
        self.logger.info('Search for Assets - Asset Manager: %s', asset_manager_id)
        search_params = {}

        if asset_ids:
            search_params['asset_ids'] = ','.join(asset_ids)
        if asset_classes:
            search_params['asset_classes'] = ','.join(asset_classes)
        if asset_types:
            search_params['asset_types'] = ','.join(asset_types)
        if fields:
            search_params['fields'] = ','.join(fields)
        if page_no is not None:
            search_params['page_no'] = page_no
        if page_size:
            search_params['page_size'] = page_size

        url = '%s/assets/%s' % (self.endpoint, asset_manager_id)
        response = self.session.get(url, params=search_params)
        if response.ok:
            asset_dicts = response.json()
            self.logger.info('Returned %s Assets.', len(asset_dicts))
            return asset_dicts
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def assets_by_asset_manager(self, asset_manager_id):
        self.logger.info('Retrieve Assets By Asset Manager: %s', asset_manager_id)
        url = '%s/assets/%s' % (self.endpoint, asset_manager_id)
        response = self.session.get(url)
        if response.ok:
            assets = [json_to_asset(json_asset) for json_asset in response.json()]
            self.logger.info('Returned %s Assets.', len(assets))
            return assets
        else:
            self.logger.error(response.text)
            response.raise_for_status()
    
    def assets_lifecycle(self, asset_manager_id, business_date, asset_ids):
        self.logger.info('Retrieve Assets Lifecycle. Asset Manager: %s', asset_manager_id)
        url = '%s/asset-lifecycle/%s' % (self.endpoint, asset_manager_id)
        params = {'business_date': business_date.isoformat(),
                  'asset_ids': ','.join(asset_ids)}
        response = self.session.get(url, params=params)
        if response.ok:
            asset_lifecycles = response.json()
            self.logger.info('Returned %s Asset Lifecycles.', len(asset_lifecycles))
            return asset_lifecycles
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def clear(self, asset_manager_id):
        """ This method deletes all the data for an asset_manager_id.
            It should be used with extreme caution.  In production it
            is almost always better to Inactivate rather than delete. """
        self.logger.info('Clear Assets - Asset Manager: %s', asset_manager_id)
        url = '%s/clear/%s' % (self.endpoint, asset_manager_id)
        response = self.session.delete(url)
        if response.ok:
            count = response.json().get('count', 'Unknown')
            self.logger.info('Deleted %s Assets.', count)
            return count
        else:
            self.logger.error(response.text)
            response.raise_for_status()



import csv
from datetime import date, datetime
from dateutil.parser import parse
from decimal import Decimal
from amaasutils.csv_upload_utils import *

from amaascore.tools.csv_tools import csv_stream_to_objects
from amaasutils.logging_utils import DEFAULT_LOGGING

from amaascore.assets.interface import AssetsInterface
from amaascore.parties.interface import PartiesInterface
from amaascore.books.interface import BooksInterface
from amaascore.corporate_actions.interface import CorporateActionsInterface
from amaascore.market_data.interface import MarketDataInterface
from amaascore.transactions.interface import TransactionsInterface
from amaascore.asset_managers.interface import AssetManagersInterface

from amaascore import parties
from amaascore import assets
from amaascore import transactions

from amaascore.assets.children import Link, Reference
from amaascore.parties.children import Address, Email, Link, Reference, Comment
from amaascore.transactions.children import Charge, Code, Comment, Link, Party, Rate, Reference

CHILDREN_CLASS = {'asset': {'links': assets.children.Link, 'references': assets.children.Reference},
                  'party': {'addresses': parties.children.Address, 'emails': parties.children.Email,
                            'links': parties.children.Link, 'references': parties.children.Reference,
                            'comments': parties.children.Comment},
                  'transaction': {'charges': transactions.children.Charge, 'codes': transactions.children.Code,
                                  'comments': transactions.children.Comment, 'links': transactions.children.Link,
                                  'rates': transactions.children.Rate, 'parties': transactions.children.Party,
                                  'references': transactions.children.Reference},
                  'asset_manager': {},
                  'book': {},
                  'corporate_action': {'references': assets.children.Reference}} #no Reference class found in corporate_actions

def direct_to_class(amaasclass):
    """direct from amaasclass (first params given in the row) to the dictionary of the children class"""
    if amaasclass in ASSET:
        return CHILDREN_CLASS['asset']
    elif amaasclass in PARTY:
        return CHILDREN_CLASS['party']
    elif amaasclass in TRANSACTION:
        return CHILDREN_CLASS['transaction']
    elif amaasclass in BOOK:
        return CHILDREN_CLASS['book']
    elif amaasclass in CORPORATE_ACTION:
        return CHILDREN_CLASS['corporate_action']
    elif amaasclass in MARKET_DATA:
        return None #None for now
    else:
        return CHILDREN_CLASS['asset_manager']

def interface_direct_class(data_class):
    """help to direct to the correct interface interacting with DB by class name only"""
    if data_class in ASSET:
        interface = AssetsInterface()
    elif data_class in PARTY:
        interface = PartiesInterface()
    elif data_class in BOOK:
        interface = BooksInterface()
    elif data_class in CORPORATE_ACTION:
        interface = CorporateActionsInterface()
    elif data_class in MARKET_DATA:
        interface = MarketDataInterface()
    elif data_class in TRANSACTION:
        interface = TransactionsInterface()
    else:
        interface = AssetManagersInterface()
    return interface

def interface_direct_csvpath(csvpath):
    """help to direct to the correct interface interacting with DB by csvfile path"""
    with open(csvpath) as csvfile:
        reader = csv.DictReader(csvfile)
        for row in reader:
            data_class = row.pop('amaasclass', '')
            return interface_direct_class(data_class)

def process_normal(_dict):
    """
    this method process the _dict to correct dict to be called by class constructor
    this method will be imported and called by main csv uploader function
    """
    cooked_dict = group_raw_to_formatted_string_dict(_dict)
    data_class = cooked_dict.pop('amaasclass', '')
    children_class_dict = direct_to_class(data_class)
    tasty_dict = dict()
    for cooked_key, cooked_value in cooked_dict.items():
        if cooked_key in CHILDREN_SIGNAL:
            processed_dict = {cooked_key: formatted_string_to_others(cooked_value, children_class_dict[cooked_key])}
        elif cooked_key == 'links':
            processed_dict = {cooked_key: formatted_string_to_links(cooked_value, children_class_dict[cooked_key])}
        else:
            processed_dict = {cooked_key: process_value_with_header(cooked_key, cooked_value)}
        tasty_dict.update(processed_dict)
    return tasty_dict



from __future__ import absolute_import, division, print_function, unicode_literals

from datetime import datetime, date
from dateutil.parser import parse
from decimal import Decimal
import sys

from amaascore.assets.asset import Asset

# This extremely ugly hack is due to the whole Python 2 vs 3 debacle.
type_check = str if sys.version_info >= (3, 0, 0) else (str, unicode)


class Fund(Asset):

    def __init__(self, asset_manager_id, asset_id, fund_type, nav=None, expense_ratio=None, net_assets=None,
                 asset_issuer_id=None, asset_status='Active', roll_price=False, display_name='', description='',
                 country_id=None, venue_id=None, currency=None, creation_date=None,
                 links=None, references=None,
                 *args, **kwargs):
        if not hasattr(self, 'asset_class'):  # A more specific child class may have already set this
            self.asset_class = 'Fund'
        self.fund_type = fund_type
        self.creation_date = creation_date
        self.nav = nav
        self.expense_ratio = expense_ratio
        self.net_assets = net_assets
        super(Fund, self).__init__(asset_manager_id=asset_manager_id, asset_id=asset_id, fungible=True,
                                   asset_issuer_id=asset_issuer_id, asset_status=asset_status,
                                   display_name=display_name, roll_price=roll_price,
                                   description=description, country_id=country_id, venue_id=venue_id,
                                   currency=currency, links=links, references=references, *args, **kwargs)

    #  Perhaps this comes from the linked Fund
    @property
    def creation_date(self):
        if hasattr(self, '_creation_date'):
            return self._creation_date

    @creation_date.setter
    def creation_date(self, value):
        """
        The date on which the bond was issued.
        :param creation_date:
        :return:
        """
        self._creation_date = parse(value).date() if isinstance(value, type_check) else value

    @property
    def nav(self):
        if hasattr(self, '_nav'):
            return self._nav

    @nav.setter
    def nav(self, nav):
        """

        :param nav:
        :return:
        """
        self._nav = Decimal(nav) if nav else None

    @property
    def expense_ratio(self):
        if hasattr(self, '_expense_ratio'):
            return self._expense_ratio

    @expense_ratio.setter
    def expense_ratio(self, expense_ratio):
        """

        :param expense_ratio:
        :return:
        """
        self._expense_ratio = Decimal(expense_ratio) if expense_ratio else None

    @property
    def net_assets(self):
        if hasattr(self, '_net_assets'):
            return self._net_assets

    @net_assets.setter
    def net_assets(self, net_assets):
        """

        :param net_assets: An integer representing the net assets of the fund
        :return:
        """
        self._net_assets = net_assets

    @property
    def fund_type(self):
        return self._fund_type

    @fund_type.setter
    def fund_type(self, fund_type):
        """

        :param fund_type: One of ['Open', 'Closed', 'ETF']
        :return:
        """
        if fund_type in ['Open', 'Closed', 'ETF']:
            self._fund_type = fund_type



from __future__ import absolute_import, division, print_function, unicode_literals

from datetime import date, datetime
from dateutil.parser import parse
import sys

from amaascore.assets.derivative import Derivative
from amaascore.assets.option_mixin import OptionMixin

# This extremely ugly hack is due to the whole Python 2 vs 3 debacle.
type_check = str if sys.version_info >= (3, 0, 0) else (str, unicode)


class BondOption(Derivative, OptionMixin):

    def __init__(self, asset_manager_id, option_type, strike, underlying_asset_id, option_style, asset_id=None,
                 asset_issuer_id=None, asset_status='Active', display_name='', description='', country_id=None,
                 currency=None, venue_id=None, issue_date=date.min, expiry_date=date.max,
                 links=None, references=None, *args, **kwargs):
        self.option_type = option_type
        self.strike = strike
        self.underlying_asset_id = underlying_asset_id
        self.option_style = option_style
        self.expiry_date = expiry_date
        super(BondOption, self).__init__(asset_manager_id=asset_manager_id, asset_id=asset_id,
                                         asset_issuer_id=asset_issuer_id, asset_status=asset_status,
                                         display_name=display_name, currency=currency,
                                         description=description, country_id=country_id, venue_id=venue_id,
                                         links=links, references=references,
                                         issue_date=issue_date, *args, **kwargs)

    @property
    def expiry_date(self):
        if hasattr(self, '_expiry_date'):
            return self._expiry_date

    @expiry_date.setter
    def expiry_date(self, value):
        """
        The date on which the Futures contract expires
        :param expiry_date:
        :return:
        """
        if value:
            self._expiry_date = parse(value).date() if isinstance(value, type_check) else value



from __future__ import absolute_import, division, print_function, unicode_literals

import datetime
from decimal import Decimal

from amaascore.core.amaas_model import AMaaSModel


class Position(AMaaSModel):

    def __init__(self, asset_manager_id, book_id, account_id, accounting_type,
                 asset_id, quantity, average_price=None, client_id=None, *args, **kwargs):
        self.client_id = client_id
        self.asset_manager_id = asset_manager_id
        self.book_id = book_id
        self.account_id = account_id
        self.accounting_type = accounting_type
        self.asset_id = asset_id
        self.quantity = quantity
        self.average_price = average_price
        super(Position, self).__init__(*args, **kwargs)

    @property
    def quantity(self):
        return self._quantity

    @quantity.setter
    def quantity(self, value):
        """
        Force the quantity to always be a decimal
        :param value:
        :return:
        """
        self._quantity = Decimal(value)

    @property
    def average_price(self):
        if hasattr(self, '_average_price'):
            return self._average_price

    @average_price.setter
    def average_price(self, value):
        """
        Force the average_price to always be a decimal
        :param value:
        :return:
        """
        if value is not None:
            self._average_price = Decimal(value)


from __future__ import absolute_import, division, print_function, unicode_literals

from amaascore.core.amaas_model import AMaaSModel


class Reference(AMaaSModel):

    def __init__(self, reference_value, reference_primary=False, *args, **kwargs):
        self.reference_value = reference_value
        self.reference_primary = reference_primary
        super(Reference, self).__init__(*args, **kwargs)



from datetime import datetime, date
from dateutil.parser import parse
from decimal import Decimal
import sys

from amaascore.assets.asset import Asset
from amaascore.assets.enums import PRIVATE_INVESTMENT_CATEGORY, PRIVATE_INVESTMENT_SHARE_TYPE,\
    PRIVATE_INVESTMENT_SUBCATEGORY

# This extremely ugly hack is due to the whole Python 2 vs 3 debacle.
type_check = str if sys.version_info >= (3, 0, 0) else (str, unicode)


class PrivateInvestment(Asset):

    def __init__(self, asset_manager_id, asset_id, client_id, asset_issuer_id=None,
                 asset_status='Active', display_name='', roll_price=True,
                 description='', country_id=None, venue_id=None, currency=None, additional=None,
                 comments=None, links=None, references=None,
                 category=None, sub_category=None, investment_date=None, num_shares=None,
                 price_share=None, share_class=None, series=None, share_type=None, coupon=None,
                 coupon_freq=None,
                 upfront_fee=None, exit_fee=None, management_fee=None, performance_fee=None,
                 hurdle=None, margin=None, high_water_mark=None, maturity_date=None,
                 lock_up_period=None, investment_term=None,
                 *args, **kwargs):
        if not hasattr(self, 'asset_class'):  # More specific child class may have already set this
            self.asset_class = 'PrivateInvestment'
        self.category = category
        self.sub_category = sub_category
        self.investment_date = investment_date
        self.num_shares = num_shares
        self.price_share = price_share
        self.share_class = share_class
        self.series = series
        self.share_type = share_type
        self.coupon = coupon
        self.coupon_freq = coupon_freq
        self.upfront_fee = upfront_fee  # These fees should probably be on the Transaction.  TODO.
        self.exit_fee = exit_fee  # These fees should probably be on the Transaction.  TODO.
        self.management_fee = management_fee  # These fees should probably be on the Transaction.  TODO.
        self.performance_fee = performance_fee  # These fees should probably be on the Transaction.  TODO.
        self.hurdle = hurdle
        self.margin = margin
        self.high_water_mark = high_water_mark
        self.maturity_date = maturity_date
        self.lock_up_period = lock_up_period
        self.investment_term = investment_term
        super(PrivateInvestment, self).__init__(asset_manager_id=asset_manager_id, asset_id=asset_id,
                                                fungible=False, asset_issuer_id=asset_issuer_id,
                                                asset_status=asset_status, display_name=display_name,
                                                roll_price=roll_price, description=description,
                                                country_id=country_id, venue_id=venue_id,
                                                currency=currency,
                                                comments=comments, links=links, references=references,
                                                client_id=client_id, additional=additional, *args, **kwargs)

    @property
    def category(self):
        return self._category

    @category.setter
    def category(self, category):
        if category in PRIVATE_INVESTMENT_CATEGORY:
            self._category=category
        else:
            raise ValueError('Invalid input of category, please indicate Others if %s not in our list' % category)

    @property
    def sub_category(self):
        return self._sub_category

    @sub_category.setter
    def sub_category(self, sub_category):
        category = self._category
        if category in PRIVATE_INVESTMENT_SUBCATEGORY.keys():
            if sub_category in PRIVATE_INVESTMENT_SUBCATEGORY[category]:
                self._sub_category = sub_category
            else:
                raise ValueError('Invalid input of sub_category: %s' % sub_category)
        else:
            raise ValueError('please set up category correctly')

    @property
    def investment_date(self):
        return self._investment_date

    @investment_date.setter
    def investment_date(self, investment_date):
        if investment_date:
            self._investment_date = parse(investment_date).date() if isinstance(investment_date, type_check)\
                else investment_date

    @property
    def num_shares(self):
        return self._num_shares

    @num_shares.setter
    def num_shares(self, num_shares):
        if isinstance(num_shares, (str, int)):
            self._num_shares = int(num_shares)
        else:
            raise ValueError("num_shares should be an integer :%s" % num_shares)

    @property
    def price_share(self):
        return self._price_share

    @price_share.setter
    def price_share(self, price_share):
        if price_share:
            self._price_share = Decimal(price_share)

    @property
    def share_class(self):
        return self._share_class

    @share_class.setter
    def share_class(self, share_class):
        self._share_class = share_class

    @property
    def share_type(self):
        return self._share_type

    @share_type.setter
    def share_type(self, share_type):
        if share_type in PRIVATE_INVESTMENT_SHARE_TYPE:
            self._share_type = share_type
        else:
            raise ValueError('Invalid input of share_type %s not in our list' % share_type)

    @property
    def maturity_date(self):
        return self._maturity_date

    @maturity_date.setter
    def maturity_date(self, maturity_date):
        if maturity_date:
            self._maturity_date = parse(maturity_date).date() if isinstance(maturity_date, type_check)\
                else maturity_date

    @property
    def lock_up_period(self):
        return self._lock_up_period

    @lock_up_period.setter
    def lock_up_period(self, lock_up_period):
        """ This lockup period is in months.  This might change to a relative delta."""
        try:
            if isinstance(lock_up_period, (str, int)):
                self._lock_up_period = int(lock_up_period)
        except Exception:
            raise ValueError('invalid input of lock up period %s, cannot be converted to an int' %
                             lock_up_period)
    @property
    def investment_term(self):
        return self._investment_term

    @investment_term.setter
    def investment_term(self, investment_term):
        """ This investment term is in months.  This might change to a relative delta."""
        try:
            if isinstance(investment_term, (str, int)):
                self._investment_term = int(investment_term)
        except Exception:
            raise ValueError('invalid input of investment type %s, cannot be converted to an int' %
                             investment_term)



from __future__ import absolute_import, division, print_function, unicode_literals

import json
import logging

from amaascore.config import ENVIRONMENT
from amaascore.core.amaas_model import json_handler
from amaascore.core.interface import Interface
from amaascore.parties.utils import json_to_party


class PartiesInterface(Interface):

    def __init__(self, environment=ENVIRONMENT, logger=None, endpoint=None, username=None, password=None):
        self.logger = logger or logging.getLogger(__name__)
        super(PartiesInterface, self).__init__(endpoint=endpoint, 
                                               endpoint_type='parties', 
                                               environment=environment, 
                                               username=None, 
                                               password=None)

    def new(self, party):
        self.logger.info('New Party - Asset Manager: %s - Party ID: %s', party.asset_manager_id, party.party_id)
        url = '%s/parties/%s' % (self.endpoint, party.asset_manager_id)
        response = self.session.post(url, json=party.to_interface())
        if response.ok:
            party = json_to_party(response.json())
            return party
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def amend(self, party):
        self.logger.info('Amend Party - Asset Manager: %s - Party ID: %s', party.asset_manager_id, party.party_id)
        url = '%s/parties/%s/%s' % (self.endpoint, party.asset_manager_id, party.party_id)
        response = self.session.put(url, json=party.to_interface())
        if response.ok:
            party = json_to_party(response.json())
            return party
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def partial(self, asset_manager_id, party_id, updates):
        self.logger.info('Partial Amend Asset - Asset Manager: %s - Party ID: %s', asset_manager_id,
                         party_id)
        url = '%s/parties/%s/%s' % (self.endpoint, asset_manager_id, party_id)
        # Setting handler ourselves so we can be sure Decimals work
        response = self.session.patch(url, data=json.dumps(updates, default=json_handler), headers=self.json_header)
        if response.ok:
            party = json_to_party(response.json())
            return party
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def retrieve(self, asset_manager_id, party_id, version=None):
        self.logger.info('Retrieve Party - Asset Manager: %s - Party ID: %s', asset_manager_id, party_id)
        url = '%s/parties/%s/%s' % (self.endpoint, asset_manager_id, party_id)
        if version:
            url += '?version=%d' % int(version)
        response = self.session.get(url)
        if response.ok:
            return json_to_party(response.json())
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def deactivate(self, asset_manager_id, party_id):
        self.logger.info('Deactivate Party - Asset Manager: %s - Party ID: %s', asset_manager_id, party_id)
        url = '%s/parties/%s/%s' % (self.endpoint, asset_manager_id, party_id)
        json = {'party_status': 'Inactive'}
        response = self.session.patch(url, json=json)
        if response.ok:
            self.logger.info(response.text)
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def search(self, asset_manager_id, party_ids=None, party_classes=None, party_types=None, country_ids=None):
        self.logger.info('Search Parties - Asset Manager: %s', asset_manager_id)
        search_params = {}
        # Potentially roll this into a loop through args rather than explicitly named - depends on additional validation
        if party_ids:
            search_params['party_ids'] = ','.join(party_ids)
        if party_classes:
            search_params['party_classes'] = ','.join(party_classes)
        if party_types:
            search_params['party_types'] = ','.join(party_types)
        if country_ids:
            search_params['country_ids'] = ','.join(country_ids)
        url = '%s/parties/%s' % (self.endpoint, asset_manager_id)
        response = self.session.get(url, params=search_params)
        if response.ok:
            parties = [json_to_party(json_party) for json_party in response.json()]
            self.logger.info('Returned %s Parties.', len(parties))
            return parties
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def parties_by_asset_manager(self, asset_manager_id):
        self.logger.info('Retrieve Parties by Asset Manager: %s', asset_manager_id)
        url = '%s/parties/%s' % (self.endpoint, asset_manager_id)
        response = self.session.get(url)
        if response.ok:
            parties = [json_to_party(json_party) for json_party in response.json()]
            self.logger.info('Returned %s Parties.', len(parties))
            return parties
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def clear(self, asset_manager_id):
        """ This method deletes all the data for an asset_manager_id.
            It should be used with extreme caution.  In production it
            is almost always better to Inactivate rather than delete. """
        self.logger.info('Clear Parties - Asset Manager: %s', asset_manager_id)
        url = '%s/clear/%s' % (self.endpoint, asset_manager_id)
        response = self.session.delete(url)
        if response.ok:
            count = response.json().get('count', 'Unknown')
            self.logger.info('Deleted %s Parties.', count)
            return count
        else:
            self.logger.error(response.text)
            response.raise_for_status()



from __future__ import absolute_import, division, print_function, unicode_literals
import inspect

from amaascore.transactions.cash_transaction import CashTransaction
from amaascore.transactions.enums import CASH_TRANSACTION_TYPES
from amaascore.transactions.position import Position
from amaascore.transactions.transaction import Transaction
from amaascore.transactions.mtm_result import MTMResult
from amaascore.transactions.transaction_pnl import TransactionPNL
from amaascore.transactions.position_pnl import PositionPNL

def json_to_position(json_position):
    position = Position(**json_position)
    return position


def json_to_transaction(json_transaction):
    # Iterate through the Transaction children, converting the various JSON attributes into the relevant class type
    for (collection_name, clazz) in Transaction.children().items():
        children = json_transaction.pop(collection_name, {})
        collection = {}
        for (child_type, child_json) in children.items():
            # Handle the case where there are multiple children for a given type - e.g. links
            if isinstance(child_json, list):
                child = set()
                for child_json_in_list in child_json:
                    child.add(clazz(**child_json_in_list))
            else:
                child = clazz(**child_json)
            collection[child_type] = child
        json_transaction[collection_name] = collection
    transaction_type = json_transaction.get('transaction_type')
    clazz = CashTransaction if transaction_type in CASH_TRANSACTION_TYPES else Transaction
    args = inspect.getargspec(clazz.__init__)
    # Some fields are always added in, even though they're not explicitly part of the constructor
    clazz_args = args.args + clazz.amaas_model_attributes()
    # is not None is important so it includes zeros and False
    constructor_dict = {arg: json_transaction.get(arg) for arg in clazz_args
                        if json_transaction.get(arg) is not None and arg != 'self'}
    transaction = clazz(**constructor_dict)
    return transaction


def json_to_mtm_result(mtm_result_json):
    args = inspect.getfullargspec(MTMResult.__init__)
    mandatory = set(args.args[1:len(args.args) - len(args.defaults)])
    missing = mandatory - \
        set([attr for attr in mandatory if mtm_result_json.get(attr) is not None])
    if not missing:
        return MTMResult(**mtm_result_json)
    else:
        raise ValueError("Missing Fields: %s in class: MTMResult" %
                         ",".join(missing))

    
def json_to_transaction_pnl(transaction_pnl_json):
    args = inspect.getfullargspec(TransactionPNL.__init__)
    mandatory = set(args.args[1:len(args.args) - len(args.defaults)])
    missing = mandatory - \
        set([attr for attr in mandatory if transaction_pnl_json.get(attr) is not None])
    if not missing:
        return TransactionPNL(**transaction_pnl_json)
    else:
        raise ValueError("Missing Fields: %s in class: TransactionPNL" %
                         ",".join(missing))


def json_to_position_pnl(position_pnl_json):
    args = inspect.getfullargspec(PositionPNL.__init__)
    mandatory = set(args.args[1:len(args.args) - len(args.defaults)])
    missing = mandatory - \
        set([attr for attr in mandatory if position_pnl_json.get(attr) is not None])
    if not missing:
        return PositionPNL(**position_pnl_json)
    else:
        raise ValueError("Missing Fields: %s in class: PositionPNL" %
                         ",".join(missing))


from __future__ import absolute_import, division, print_function, unicode_literals

from datetime import date
from dateutil.parser import parse
import re
import sys
import uuid
import inspect
import json

from amaascore.core.amaas_model import to_json_string
from amaascore.assets.children import Link
from amaascore.core.amaas_model import AMaaSModel
from amaascore.core.comment import Comment
from amaascore.core.reference import Reference
from amaasutils.hash import compute_hash


# This extremely ugly hack is due to the whole Python 2 vs 3 debacle.
type_check = str if sys.version_info >= (3, 0, 0) else (str, unicode)


class Asset(AMaaSModel):

    @staticmethod
    def pricing_method():
        return 'Derived'

    @staticmethod
    def children():
        return {'comments': Comment, 'links': Link, 'references': Reference}

    def __init__(self, asset_manager_id, fungible, asset_issuer_id=None, asset_id=None, asset_status='Active',
                 country_id=None, venue_id=None, currency=None, issue_date=date.min,
                 roll_price=True, display_name='', description='', maturity_date=date.max,
                 comments=None, links=None, references=None, client_additional=None,
                 *args, **kwargs):
        self.asset_manager_id = asset_manager_id
        self.asset_id = asset_id or uuid.uuid4().hex
        if not hasattr(self, 'asset_class'):  # A more specific child class may have already set this
            self.asset_class = 'Asset'
        # This can be overridden if necessary but I think it probably shouldn't
        self.asset_type_display = ' '.join(re.findall('[A-Z][^A-Z]*', self.__class__.__name__))
        self.asset_type = self.__class__.__name__
        self.fungible = fungible
        self.asset_issuer_id = asset_issuer_id
        self.asset_status = asset_status
        self.country_id = country_id
        self.venue_id = venue_id
        self.currency = currency
        self.issue_date = issue_date
        self.maturity_date = maturity_date
        self.roll_price = roll_price
        self.display_name = display_name
        self.description = description
        self.client_additional = client_additional  # A field to allow people to build their own assets
        # Defaults are here not in constructor for mutability reasons.
        self.comments = comments.copy() if comments else {}
        self.links = links.copy() if links else {}
        self.references = references.copy() if references else {}
        self.references['Argomi'] = Reference(reference_value=self.asset_id)  # Upserts the Argomi Reference
        super(Asset, self).__init__(*args, **kwargs)

    def reference_types(self):
        """
        TODO - are these helper functions useful?
        :return:
        """
        return self.references.keys()

    @property
    def issue_date(self):
        if hasattr(self, '_issue_date'):
            return self._issue_date

    @issue_date.setter
    def issue_date(self, value):
        """
        The date on which the asset is issued
        :param value:
        :return:
        """
        self._issue_date = parse(value).date() if isinstance(value, type_check) else value

    @property
    def maturity_date(self):
        if hasattr(self, '_maturity_date'):
            return self._maturity_date

    @maturity_date.setter
    def maturity_date(self, value):
        """
        The date on which the asset matures and no longer holds value
        :param value:
        :return:
        """
        self._maturity_date = parse(value).date() if isinstance(value, type_check) else value

    @property
    def roll_price(self):
        if hasattr(self, '_roll_price'):
            return self._roll_price

    @roll_price.setter
    def roll_price(self, value):
        """
        Always convert to bool if the service/database returns 0 or 1
        """
        if value is not None:
            self._roll_price = True if value else False

    @property
    def fungible(self):
        if hasattr(self, '_fungible'):
            return self._fungible

    @fungible.setter
    def fungible(self, value):
        """
        Always convert to bool if the service/database returns 0 or 1
        """
        if value is not None:
            self._fungible = True if value else False

    def __str__(self):
        return "Asset object - ID: %s" % self.asset_id

    def get_country_codes(self):
        return [self.country_id]

    def country_and_venue_codes(self):
        """
        This function returns dictionary in the format of {"country_codes": [XYZ, ABC], "venue_id": [ABC]}
        Where country_codes = self.country_id and venue_id = self.venue_id
        except for FX where we also return the country_ids associated with the legs of the FX transaction
        (e.g. USDJPY returns USA + JPN).
        """
        codes = {'country_codes': self.get_country_codes(), 'venue_id': [self.venue_id]}
        return codes

    def get_currencies(self):
        return [self.currency]

    @property
    def hashcode(self):
        asset_dict = {key: getattr(self, key) for key in self.to_dict() if key != 'hashcode'}
        type_args = inspect.signature(self.__init__).parameters.keys()
        ignored_attributes = [key for key in asset_dict if key not in type_args]
        data = json.loads(to_json_string(asset_dict))
        return compute_hash(data, ignored_attributes=ignored_attributes)



from __future__ import absolute_import, division, print_function, unicode_literals

import datetime
from decimal import Decimal
import json
import sys

# This extremely ugly hack is due to the whole Python 2 vs 3 debacle.
type_check = str if sys.version_info >= (3, 0, 0) else (str, unicode)


def json_handler(value):
    if isinstance(value, (datetime.datetime, datetime.date)):
        return value.isoformat()
    if isinstance(value, datetime.timedelta):
        return (datetime.datetime.min + value).time().isoformat()
    if isinstance(value, set):
        return list(value)
    if isinstance(value, Decimal):
        return str(value)
    if isinstance(value, AMaaSModel):
        return value.to_json()
    raise TypeError("JSON Handler Failed on value '%s': Unknown type '%s'" % (value, type(value)))


def to_json(dict_to_convert):
    return json.loads(to_json_string(dict_to_convert))


def to_json_string(dict_to_convert):
    return json.dumps(dict_to_convert, ensure_ascii=False, default=json_handler, indent=4, separators=(',', ': '))


class AMaaSModel(object):

    @staticmethod
    def non_interface_attributes():
        """ Potentially convert this to attribute annotations """
        return []

    @staticmethod
    def amaas_model_attributes():
        return ['created_by', 'updated_by', 'created_time', 'updated_time', 'version']


    def __init__(self, *args, **kwargs):
        self.version = kwargs.get('version') or 1
        self.created_by = kwargs.get('created_by')
        self.updated_by = kwargs.get('updated_by')
        self.created_time = kwargs.get('created_time')  # Comes from database
        self.updated_time = kwargs.get('updated_time')  # Comes from database

    @property
    def version(self):
        return self._version

    @version.setter
    def version(self, version):
        """ Cast string versions to int (if read from a file etc) """
        if isinstance(version, type_check):
            self._version = int(version)
        elif isinstance(version, int):
            self._version = version

    def to_interface(self):
        """
        Returns only the JSON attributes required when interfacing with the AMaaS Core services.
        Non-interface attributes are popped out.
        :return:
        """
        dict_to_convert = self.__dict__
        [dict_to_convert.pop(attr) for attr in self.non_interface_attributes()]
        return self.to_json(dict_to_convert)

    def to_dict(self, dict_to_convert=None):
        dict_to_convert = dict_to_convert or self.__dict__
        # Convert internal property values (_XYZ) to the correctly named one (XYZ)
        # Is there a better way of doing this instead of relying on the first character?
        mapped_dict = {}
        for key, value in dict_to_convert.items():
            key = key[1:] if key[0] == '_' else key
            mapped_dict[key] = value
        return mapped_dict

    def to_json(self, dict_to_convert=None):
        return to_json(self.to_dict(dict_to_convert=dict_to_convert))

    def to_json_string(self, dict_to_convert=None):
        return to_json_string(self.to_dict(dict_to_convert=dict_to_convert))

    def __repr__(self):
        """
        # TODO - check the nested dictionaries
        :return:
        """
        return str(self.to_dict())

    def __eq__(self, other):
        """Override the default Equals behavior"""
        if isinstance(other, self.__class__):
            my_dict = self.to_dict()
            other_dict = other.to_dict()
            # Strip out the database generated fields:
            [my_dict.pop(attr, None) for attr in self.amaas_model_attributes()]
            [other_dict.pop(attr, None) for attr in self.amaas_model_attributes()]

            return my_dict == other_dict
        return NotImplemented

    def __ne__(self, other):
        if isinstance(other, self.__class__):
            return not self.__eq__(other)
        return NotImplemented

    def __hash__(self):
        """Override the default hash behavior (that returns the id or the object)"""
        output = []
        for (key, value) in self.to_dict().items():
            # Remove the internal attributes since they shouldn't be used for ordering etc
            if key not in self.amaas_model_attributes():
                output_value = hash(tuple(sorted(value))) if isinstance(value, dict) else value
                output.append((key, output_value))
        return hash(tuple(sorted(output)))




from __future__ import absolute_import, division, print_function, unicode_literals

from amaascore.core.amaas_model import AMaaSModel

class Domain(AMaaSModel):

    def __init__(self, asset_manager_id, domain, is_primary=True, domain_status='Active',
                 *args, **kwargs):
        self.asset_manager_id = asset_manager_id
        self.domain = domain
        self.is_primary = is_primary
        self.domain_status = domain_status
        super(Domain, self).__init__(*args, **kwargs)



from __future__ import absolute_import, division, print_function, unicode_literals

from amaascore.corporate_actions.corporate_action import CorporateAction


class Split(CorporateAction):

    def __init__(self, asset_manager_id, corporate_action_id, record_date, ratio, corporate_action_status='Open',
                 asset_id=None, party_id=None, declared_date=None, settlement_date=None, elective=False, message=None,
                 description='', references=None, *args, **kwargs):
        self.ratio = ratio
        super(Split, self).__init__(asset_manager_id=asset_manager_id, corporate_action_id=corporate_action_id,
                                    record_date=record_date, corporate_action_status=corporate_action_status,
                                    asset_id=asset_id, party_id=party_id, declared_date=declared_date,
                                    settlement_date=settlement_date, elective=elective, message=message,
                                    description=description, references=references, *args, **kwargs)

    @property
    def ratio(self):
        if hasattr(self, '_ratio'):
            return self._ratio

    @ratio.setter
    def ratio(self, ratio):
        """
        The split ratio of the corporate action - i.e. the ratio of new shares to old shares
        :param ratio: A tuple representing (original_count, new_count).  For example (1, 2) is a doubling stock split.
        (3, 1) is a 3:1 reverse stock split.
        :return:
        """
        if isinstance(ratio, tuple):
            self._ratio = ratio
        else:
            raise TypeError('Invalid ratio type: %s' % type(ratio))



from __future__ import absolute_import, division, print_function, unicode_literals

import inspect

#  All possible class names must be inserted into the globals collection.
#  If there is a better way of doing this, please suggest!
from amaascore.parties.broker import Broker
from amaascore.parties.company import Company
from amaascore.parties.asset_manager import AssetManager
from amaascore.parties.exchange import Exchange
from amaascore.parties.fund import Fund
from amaascore.parties.government_agency import GovernmentAgency
from amaascore.parties.individual import Individual
from amaascore.parties.organisation import Organisation
from amaascore.parties.party import Party
from amaascore.parties.sub_fund import SubFund


def json_to_party(json_to_convert):
    # Iterate through the party children, converting the various JSON attributes into the relevant class type
    for (collection_name, clazz) in Party.children().items():
        children = json_to_convert.pop(collection_name, {})
        collection = {}
        for (child_type, child_json) in children.items():
            # Handle the case where there are multiple children for a given type - e.g. links
            if isinstance(child_json, list):
                child = set()
                for child_json_in_list in child_json:
                    child.add(clazz(**child_json_in_list))
            else:
                child = clazz(**child_json)
            collection[child_type] = child
        json_to_convert[collection_name] = collection
    clazz = globals().get(json_to_convert.get('party_type'))
    if not clazz:
        raise ValueError('Missing Party Type: %s' % json_to_convert.get('party_type'))
    args = inspect.getargspec(clazz.__init__)
    # Some fields are always added in, even though they're not explicitly part of the constructor
    clazz_args = args.args + clazz.amaas_model_attributes()
    # is not None is important so it includes zeros and False
    constructor_dict = {arg: json_to_convert.get(arg) for arg in clazz_args
                        if json_to_convert.get(arg) is not None and arg != 'self'}
    party = clazz(**constructor_dict)
    return party



from decimal import Decimal
from amaascore.core.amaas_model import AMaaSModel


class PositionPNL(AMaaSModel):

    def __init__(self, asset_manager_id, book_id, asset_id, period, 
                 business_date, pnl_timestamp, pnl_status='Active',
                 total_pnl=None, asset_pnl=None, fx_pnl=None, 
                 unrealised_pnl=None, realised_pnl=None, quantity=None,
                 currency=None, message=None, *args, **kwargs):
        self.asset_manager_id = asset_manager_id
        self.asset_id = asset_id
        self.book_id = book_id
        self.period = period
        self.business_date = business_date
        self.realised_pnl = realised_pnl
        self.unrealised_pnl = unrealised_pnl
        self.total_pnl = total_pnl
        self.asset_pnl = asset_pnl
        self.fx_pnl = fx_pnl
        self.pnl_status = pnl_status
        self.quantity = quantity
        self.message = message
        self.currency = currency
        self.pnl_timestamp = pnl_timestamp

        super(PositionPNL, self).__init__(*args, **kwargs)

    @property
    def quantity(self):
        return self._quantity

    @quantity.setter
    def quantity(self, val):
        if not isinstance(val, Decimal) and val is not None:
            self._quantity = Decimal(val)
        else:
            self._quantity = val        

    @property
    def period(self):
        return self._period

    @period.setter
    def period(self, val):
        if val not in ['YTD', 'MTD', 'DTD']:
            raise ValueError("""Unrecognized PnL period %s, expect
                    period to be one of the following: 'YTD', 'MTD', 'DTD'""" % str(val))
        self._period = val

    @property
    def total_pnl(self):
        return self._total_pnl

    @total_pnl.setter
    def total_pnl(self, val):
        if not isinstance(val, Decimal) and val is not None:
            self._total_pnl = Decimal(val)
        else:
            self._total_pnl = val

    @property
    def fx_pnl(self):
        return self._fx_pnl

    @fx_pnl.setter
    def fx_pnl(self, val):
        if not isinstance(val, Decimal) and val is not None:
            self._fx_pnl = Decimal(val)
        else:
            self._fx_pnl = val

    @property
    def asset_pnl(self):
        return self._asset_pnl

    @asset_pnl.setter
    def asset_pnl(self, val):
        if not isinstance(val, Decimal) and val is not None:
            self._asset_pnl = Decimal(val)
        else:
            self._asset_pnl = val



from __future__ import absolute_import, division, print_function, unicode_literals


from amaasutils.random_utils import random_string, random_decimal, random_date
from datetime import date, timedelta
from decimal import Decimal
import random

from amaascore.assets.asset import Asset
from amaascore.assets.bond import BondGovernment
from amaascore.assets.bond_option import BondOption
from amaascore.assets.currency import Currency
from amaascore.assets.cryptocurrency import Cryptocurrency
from amaascore.assets.enums import CRYPTOCURRENCY_PROOF_TYPES
from amaascore.assets.equity import Equity
from amaascore.assets.foreign_exchange import ForeignExchange, ForeignExchangeForward, ForeignExchangeSpot
from amaascore.assets.fund import Fund
from amaascore.assets.future import Future
from amaascore.assets.fx_option import ForeignExchangeOption
from amaascore.assets.sukuk import Sukuk
from amaascore.assets.synthetic import Synthetic
from amaascore.assets.private_investment import PrivateInvestment
from amaascore.assets.automobile import Automobile
from amaascore.assets.warrants import Warrant
from amaascore.core.reference import Reference

REFERENCE_TYPES = ['External']


def generate_common(asset_manager_id=None, asset_id=None, display_name=None):

    common = {'asset_manager_id': asset_manager_id or random.randint(1, 1000),
              'asset_id': asset_id or str(random.randint(1, 1000)),
              'currency': random.choice(['SGD', 'USD']),
              'display_name': display_name or random_string(10)
             }
    return common


def generate_asset(asset_manager_id=None, asset_id=None, fungible=None, roll_price=None, country_id=None):

    common = generate_common(asset_manager_id=asset_manager_id, asset_id=asset_id)
    common['fungible'] = random.choice([True, False]) if fungible is None else fungible
    common['roll_price'] = random.choice([True, False]) if roll_price is None else roll_price
    common['country_id'] = country_id or random.choice(['SGP', 'USA', 'GBR', 'JPN'])
    asset = Asset(**common)
    references = {ref_type: Reference(reference_value=random_string(10)) for ref_type in REFERENCE_TYPES}

    asset.references.update(references)
    return asset


def generate_equity(asset_manager_id, asset_id, share_class='Common', currency='USD'):
    equity = Equity(asset_manager_id=asset_manager_id or random.randint(1, 1000),
                    asset_id=asset_id or random_string(5),
                    share_class=share_class, 
                    asset_status='Active',
                    currency=currency or 'USD')
    return equity


def generate_bond(asset_manager_id=None, asset_id=None):
    props = generate_common(asset_manager_id=asset_manager_id, asset_id=asset_id)
    bond = BondGovernment(par=Decimal('1000'),
                          pay_frequency='M',  # Need to check how we want to represent this
                          coupon=Decimal('0.05'),
                          **props)
    return bond


def generate_bond_option(asset_manager_id=None, asset_id=None, option_type=None, strike=None, option_style=None):
    props = generate_common(asset_manager_id=asset_manager_id, asset_id=asset_id)
    asset = BondOption(underlying_asset_id=random_string(10),
                       option_style=option_style or random.choice(['European', 'American']),
                       option_type=option_type or random.choice(['Put', 'Call']),
                       strike=strike or Decimal(random.uniform(99.0, 102.0)).quantize(Decimal('0.05')),
                       **props)
    return asset


def generate_currency(asset_id=None):
    asset = Currency(asset_id=asset_id or random_string(3))
    return asset


def generate_cryptocurrency(asset_id=None):
    asset = Cryptocurrency(asset_id=asset_id or random_string(3),
                           proof_type=random.choice(list(CRYPTOCURRENCY_PROOF_TYPES)))
    return asset

def generate_foreignexchange(asset_id=None):
    asset = ForeignExchange(asset_id=asset_id)
    return asset


def generate_fund(asset_manager_id=None, asset_id=None):
    props = generate_common(asset_manager_id=asset_manager_id, asset_id=asset_id)
    asset = Fund(fund_type=random.choice(['Open', 'Closed']),
                 nav=random_decimal(),
                 expense_ratio=random_decimal(),
                 net_assets=1e06*random.randint(1, 10000),
                 **props)
    return asset


def generate_future(asset_manager_id=None, asset_id=None, point_value=None, expiry_date=None):
    props = generate_common(asset_manager_id=asset_manager_id, asset_id=asset_id)
    asset = Future(settlement_type=random.choice(['Cash', 'Physical']),
                   contract_size=10000,
                   point_value=point_value or Decimal('50'),
                   tick_size=Decimal('0.01'),
                   expiry_date=expiry_date or random_date(start_year=date.today().year+1),
                   **props)
    return asset


def generate_fx_forward(asset_manager_id=None, asset_id=None, underlying=None, settlement_date=None, fixing_date=None):
    props = generate_common(asset_manager_id=asset_manager_id, asset_id=asset_id, display_name=asset_id)
    asset = ForeignExchangeForward(forward_rate=random_decimal(),
                                   settlement_date=settlement_date or random_date(start_year=2017),
                                   underlying=underlying or 'USDJPY',
                                   fixing_date=fixing_date,
                                   **props)
    return asset


def generate_fx_spot(asset_manager_id=None, asset_id=None, settlement_date=None):
    props = generate_common(asset_manager_id=asset_manager_id, asset_id=asset_id, display_name=asset_id)
    tomorrow = date.today() + timedelta(days=1)
    asset = ForeignExchangeSpot(settlement_date=settlement_date or tomorrow,
                                underlying='USDJPY',
                                **props)
    return asset


def generate_fx_option(asset_manager_id=None, asset_id=None, option_type=None, strike=None, option_style=None):
    props = generate_common(asset_manager_id=asset_manager_id, asset_id=asset_id)
    del props['currency']  # FX doesn't have a currency
    asset = ForeignExchangeOption(underlying_asset_id=random_string(10),
                                  option_style=option_style or random.choice(['European', 'American']),
                                  option_type=option_type or random.choice(['Put', 'Call']),
                                  strike=strike or Decimal(random.uniform(99.0, 102.0)).quantize(Decimal('0.05')),
                                  **props)
    return asset


def generate_sukuk(asset_manager_id=None, asset_id=None):
    props = generate_common(asset_manager_id=asset_manager_id, asset_id=asset_id)
    sukuk = Sukuk(maturity_date=random_date(end_year=2050), **props)
    return sukuk


def generate_synthetic(asset_manager_id=None, asset_id=None):
    props = generate_common(asset_manager_id=asset_manager_id, asset_id=asset_id)
    synthetic = Synthetic(**props)
    return synthetic


def generate_assets(asset_manager_ids=[], number=5):
    assets = []
    for i in range(number):
        asset = generate_asset(asset_manager_id=random.choice(asset_manager_ids))
        assets.append(asset)
    return assets

def generate_private_investment(asset_manager_id=None, asset_id=None, client_id=None):
    attributes = generate_common(asset_manager_id=asset_manager_id, asset_id=asset_id)
    """currency, display_name"""
    private_investment = PrivateInvestment(client_id=client_id or random_string(5),
                                           asset_issuer_id=random_string(8),
                                           category='Private Equity', 
                                           sub_category='Leverage Buyout Funds',
                                           num_shares=1000,
                                           price_share=1000,
                                           share_type='Ordinary Shares',
                                           maturity_date=random_date(),
                                           lock_up_period=52,
                                           investment_term=52,
                                           **attributes)
    return private_investment

def generate_automobile(asset_manager_id=None, asset_id=None, client_id=None):
    attributes = generate_common(asset_manager_id=asset_manager_id, asset_id=asset_id)
    automobile = Automobile(client_id=1, **attributes)
    return automobile

def generate_warrant(asset_manager_id=None, asset_id=None, client_id=None):
    attributes = generate_common(asset_manager_id=asset_manager_id, asset_id=asset_id)
    warrant = Warrant(client_id=1, **attributes)
    return warrant


from __future__ import absolute_import, division, print_function, unicode_literals

# TODO - Support multi-language errors
ERROR_LOOKUP = {'address_invalid': 'Invalid addresses attribute: %s. Party ID %s for Asset Manager %s',
                'address_primary': 'Must set exactly one address as primary. Party ID %s for Asset Manager %s',
                'am_type_invalid': 'Asset Manager Type: %s is invalid.  Asset Manager: %s',
                'am_account_type_invalid': 'Account Type: %s is invalid.  Asset Manager: %s',
                'book_type_invalid': 'Invalid book type %s. Book ID: %s for Asset Manager: %s',
                'country_id_invalid': 'Country ID should be a ISO 3166-1 Alpha-3 code. Value: %s',
                'currency_invalid': 'Invalid currency %s. Transaction ID: %s for asset manager: %s',
                'email_invalid': 'Invalid emails attribute: %s. Party ID %s for Asset Manager %s',
                'email_primary': 'Must set exactly one email as primary. Party ID %s for Asset Manager %s',
                'email_address_invalid': 'Invalid email: %s.',
                'amend_missing_previous': 'Cannot find party to amend: ID %s for Asset Manager %s',
                'amend_missing_attribute': 'Partial amend failed for Asset Manager: %s on party: %s - '
                                           'Attribute: %s does not exist',
                'deactivate_missing_previous': 'Cannot Deactivate Party - Cannot Find ID: %s for Asset Manager: %s',
                'party_status_invalid': 'Invalid party status %s. Party ID: %s for Asset Manager: %s',
                'transaction_action_invalid': 'Invalid transaction action %s. Transaction ID: %s for Asset Manager: %s',
                'transaction_status_invalid': 'Invalid transaction status %s. Transaction ID: %s for Asset Manager: %s',
                'transaction_type_invalid': 'Invalid transaction type %s. Transaction ID: %s for Asset Manager: %s',
                'transaction_link_not_found': 'Cannot remove link - not found'}
